import cv2
import cvlib as cv


def objects_from_cvlib(objects_list: list) -> dict:
    objects = init_default_objects()
    for key in objects:
        if key in objects_list:
            objects[key] = "yes"
    return objects


def init_default_objects():
    objects = {
        "person": "no",
        "bicycle": "no",
        "car": "no",
        "motorcycle": "no",
        "airplane": "no",
        "bus": "no",
        "train": "no",
        "truck": "no",
        "boat": "no",
        "traffic light": "no",
        "cell phone": "no",
    }
    return objects


class ObjectsMethod:
    """Base class to be inherited by all objects methods."""

    def __init__(self):
        # initialize in child class
        pass

    def analyse_image(self, subdict):
        raise NotImplementedError()


class ObjectCVLib(ObjectsMethod):
    def __init__(self, client_type=1):
        # as long as imageai is not activated this remains empty
        pass

    def detect_objects_cvlib(self, image_path):
        """Localize objects in the local image.

        Args:
            image_path: The path to the local file.
        """
        img = cv2.imread(image_path)

        _, label, _ = cv.detect_common_objects(img)
        objects = objects_from_cvlib(label)
        return objects

    def analyse_image_from_file(self, image_path):
        """Localize objects in the local image.

        Args:
            image_path: The path to the local file.
        """
        objects = self.detect_objects_cvlib(image_path)
        return objects

    def analyse_image(self, subdict):
        """Localize objects in the local image.

        Args:
            subdict: The dictionary for an image expression instance.
        """
        objects = self.analyse_image_from_file(subdict["filename"])
        for key in objects:
            subdict[key] = objects[key]

        return subdict



import glob
import os
from pandas import DataFrame, read_csv
import pooch
import importlib_resources
import collections
import random


pkg = importlib_resources.files("ammico")


def iterable(arg):
    return isinstance(arg, collections.abc.Iterable) and not isinstance(arg, str)


class DownloadResource:
    """A remote resource that needs on demand downloading.

    We use this as a wrapper to the pooch library. The wrapper registers
    each data file and allows prefetching through the CLI entry point
    ammico_prefetch_models.
    """

    # We store a list of defined resouces in a class variable, allowing
    # us prefetching from a CLI e.g. to bundle into a Docker image
    resources = []

    def __init__(self, **kwargs):
        DownloadResource.resources.append(self)
        self.kwargs = kwargs

    def get(self):
        return pooch.retrieve(**self.kwargs)


def ammico_prefetch_models():
    """Prefetch all the download resources"""
    for res in DownloadResource.resources:
        res.get()


class AnalysisMethod:
    """Base class to be inherited by all analysis methods."""

    def __init__(self, subdict: dict) -> None:
        self.subdict = subdict
        # define keys that will be set by the analysis

    def set_keys(self):
        raise NotImplementedError()

    def analyse_image(self):
        raise NotImplementedError()


def _match_pattern(path, pattern, recursive):
    # helper function for find_files
    # find all matches for a single pattern.

    if pattern.startswith("."):
        pattern = pattern[1:]
    if recursive:
        search_path = f"{path}/**/*.{pattern}"
    else:
        search_path = f"{path}/*.{pattern}"
    return list(glob.glob(search_path, recursive=recursive))


def _limit_results(results, limit):
    # helper function for find_files
    # use -1 or None to return all images
    if limit == -1 or limit is None:
        limit = len(results)

    # limit or batch the images
    if isinstance(limit, int):
        if limit < -1:
            raise ValueError("limit must be an integer greater than 0 or equal to -1")
        results = results[:limit]

    elif iterable(limit):
        if len(limit) == 2:
            results = results[limit[0] : limit[1]]
        else:
            raise ValueError(
                f"limit must be an integer or a tuple of length 2, but is {limit}"
            )
    else:
        raise ValueError(
            f"limit must be an integer or a tuple of length 2, but is {limit}"
        )

    return results


def find_files(
    path: str = None,
    pattern=["png", "jpg", "jpeg", "gif", "webp", "avif", "tiff"],
    recursive: bool = True,
    limit=20,
    random_seed: int = None,
) -> dict:
    """Find image files on the file system.

    Args:
        path (str, optional): The base directory where we are looking for the images. Defaults
            to None, which uses the ammico data directory if set or the current
            working directory otherwise.
        pattern (str|list, optional): The naming pattern that the filename should match.
                Use either '.ext' or just 'ext'
                Defaults to ["png", "jpg", "jpeg", "gif", "webp", "avif","tiff"]. Can be used to allow other patterns or to only include
                specific prefixes or suffixes.
        recursive (bool, optional): Whether to recurse into subdirectories. Default is set to True.
        limit (int/list, optional): The maximum number of images to be found.
            Provide a list or tuple of length 2 to batch the images.
            Defaults to 20. To return all images, set to None or -1.
        random_seed (int, optional): The random seed to use for shuffling the images.
            If None is provided the data will not be shuffeled. Defaults to None.
    Returns:
        dict: A nested dictionary with file ids and all filenames including the path.
    """

    if path is None:
        path = os.environ.get("AMMICO_DATA_HOME", ".")

    if isinstance(pattern, str):
        pattern = [pattern]
    results = []
    for p in pattern:
        results.extend(_match_pattern(path, p, recursive=recursive))

    if len(results) == 0:
        raise FileNotFoundError(f"No files found in {path} with pattern '{pattern}'")

    if random_seed is not None:
        random.seed(random_seed)
        random.shuffle(results)

    images = _limit_results(results, limit)

    return initialize_dict(images)


def initialize_dict(filelist: list) -> dict:
    """Initialize the nested dictionary for all the found images.

    Args:
        filelist (list): The list of files to be analyzed, including their paths.
    Returns:
        dict: The nested dictionary with all image ids and their paths."""
    mydict = {}
    for img_path in filelist:
        id_ = os.path.splitext(os.path.basename(img_path))[0]
        mydict[id_] = {"filename": img_path}
    return mydict


def check_for_missing_keys(mydict: dict) -> dict:
    """Check the nested dictionary for any missing keys in the subdicts.

    Args:
        mydict(dict): The nested dictionary with keys to check.
    Returns:
        dict: The dictionary with keys appended."""
    # check that we actually got a nested dict
    if not isinstance(mydict[next(iter(mydict))], dict):
        raise ValueError(
            "Please provide a nested dictionary - you provided {}".format(
                next(iter(mydict))
            )
        )
    # gather all existing keys of first item in a list
    subdict = mydict[next(iter(mydict))]
    if len(list(subdict.keys())) < 1:
        raise ValueError(
            "Could not get any keys to compare to - please check if your nested dict is empty!"
        )
    for key in mydict.keys():
        # compare keys of next item with first item
        if subdict.keys() != mydict[key].keys():
            # print a warning if key is not found and set to None
            keys_a = set(subdict.keys())
            keys_b = set(mydict[key].keys())
            missing_keys_in_b = keys_a - keys_b
            if missing_keys_in_b:
                print(
                    "Found missing key(s) {} in subdict {} - setting to None.".format(
                        missing_keys_in_b, key
                    )
                )
                for missing_key in missing_keys_in_b:
                    mydict[key][missing_key] = None
            # check that there are no other keys in the subdicts -
            # this would only happen if there is a key missing in the first subdict
            # then we would need to start over so best to
            # abort if this happens - this is a very unlikely case
            missing_keys_in_a = keys_b - keys_a
            if missing_keys_in_a:
                raise ValueError(
                    "Could not update missing keys - first item already missing {}".format(
                        missing_keys_in_a
                    )
                )
    return mydict


def append_data_to_dict(mydict: dict) -> dict:
    """Append entries from nested dictionaries to keys in a global dict."""

    # first initialize empty list for each key that is present
    outdict = {key: [] for key in list(mydict.values())[0].keys()}
    # now append the values to each key in a list
    for subdict in mydict.values():
        for key in subdict.keys():
            outdict[key].append(subdict[key])
    return outdict


def dump_df(mydict: dict) -> DataFrame:
    """Utility to dump the dictionary into a dataframe."""
    return DataFrame.from_dict(mydict)


def get_dataframe(mydict: dict) -> DataFrame:
    check_for_missing_keys(mydict)
    outdict = append_data_to_dict(mydict)
    return dump_df(outdict)


def is_interactive():
    """Check if we are running in an interactive environment."""
    import __main__ as main

    return not hasattr(main, "__file__")


def get_color_table():
    path_tables = pkg / "data" / "Color_tables.csv"
    df_colors = read_csv(
        path_tables,
        delimiter=";",
        dtype=str,
        encoding="UTF-8",
        header=[0, 1],
    )
    return {
        col_key: df_colors[col_key].dropna().to_dict("list")
        for col_key in df_colors.columns.levels[0]
    }



from google.cloud import vision
from google.auth.exceptions import DefaultCredentialsError
from googletrans import Translator
import spacy
import io
from ammico.utils import AnalysisMethod
import grpc
import pandas as pd
from bertopic import BERTopic
from transformers import pipeline


class TextDetector(AnalysisMethod):
    def __init__(
        self,
        subdict: dict,
        analyse_text: bool = False,
        model_names: list = None,
        revision_numbers: list = None,
    ) -> None:
        """Init text detection class.

        Args:
            subdict (dict): Dictionary containing file name/path, and possibly previous
                analysis results from other modules.
            analyse_text (bool, optional): Decide if extracted text will be further subject
                to analysis. Defaults to False.
            model_names (list, optional): Provide model names for summary, sentiment and ner
                analysis. Defaults to None, in which case the default model from transformers
                are used (as of 03/2023): "sshleifer/distilbart-cnn-12-6" (summary),
                "distilbert-base-uncased-finetuned-sst-2-english" (sentiment),
                "dbmdz/bert-large-cased-finetuned-conll03-english".
                To select other models, provide a list with three entries, the first for
                summary, second for sentiment, third for NER, with the desired model names.
                Set one of these to None to still use the default model.
            revision_numbers (list, optional): Model revision (commit) numbers on the
                Hugging Face hub. Provide this to make sure you are using the same model.
                Defaults to None, except if the default models are used; then it defaults to
                "a4f8f3e" (summary, distilbart), "af0f99b" (sentiment, distilbert),
                "f2482bf" (NER, bert).
        """
        super().__init__(subdict)
        self.subdict.update(self.set_keys())
        self.translator = Translator()
        if not isinstance(analyse_text, bool):
            raise ValueError("analyse_text needs to be set to true or false")
        self.analyse_text = analyse_text
        if self.analyse_text:
            self._initialize_spacy()
        if model_names:
            self._check_valid_models(model_names)
        if revision_numbers:
            self._check_revision_numbers(revision_numbers)
        # initialize revision numbers and models
        self._init_revision_numbers(model_names, revision_numbers)
        self._init_model(model_names)

    def _check_valid_models(self, model_names):
        # check that model_names and revision_numbers are valid lists or None
        # check that model names are a list
        if not isinstance(model_names, list):
            raise ValueError("Model names need to be provided as a list!")
        # check that enough models are provided, one for each method
        if len(model_names) != 3:
            raise ValueError(
                "Not enough or too many model names provided - three are required, one each for summary, sentiment, ner"
            )

    def _check_revision_numbers(self, revision_numbers):
        # check that revision numbers are list
        if not isinstance(revision_numbers, list):
            raise ValueError("Revision numbers need to be provided as a list!")
        # check that three revision numbers are provided, one for each method
        if len(revision_numbers) != 3:
            raise ValueError(
                "Not enough or too many revision numbers provided - three are required, one each for summary, sentiment, ner"
            )

    def _init_revision_numbers(self, model_names, revision_numbers):
        """Helper method to set the revision (version) number for each model."""
        revision_numbers_default = ["a4f8f3e", "af0f99b", "f2482bf"]
        if model_names:
            # if model_names is provided, set revision numbers for each of the methods
            # either as the provided revision number or None or as the default revision number,
            # if one of the methods uses the default model
            self._init_revision_numbers_per_model(
                model_names, revision_numbers, revision_numbers_default
            )
        else:
            # model_names was not provided, revision numbers are the default revision numbers or None
            self.revision_summary = revision_numbers_default[0]
            self.revision_sentiment = revision_numbers_default[1]
            self.revision_ner = revision_numbers_default[2]

    def _init_revision_numbers_per_model(
        self, model_names, revision_numbers, revision_numbers_default
    ):
        task_list = []
        if not revision_numbers:
            # no revision numbers for non-default models provided
            revision_numbers = [None, None, None]
        for model, revision, revision_default in zip(
            model_names, revision_numbers, revision_numbers_default
        ):
            # a model was specified for this task, set specified revision number or None
            # or: model for this task was set to None, so we take default version number for default model
            task_list.append(revision if model else revision_default)
        self.revision_summary = task_list[0]
        self.revision_sentiment = task_list[1]
        self.revision_ner = task_list[2]

    def _init_model(self, model_names):
        """Helper method to set the model name for each analysis method."""
        # assign models for each of the text analysis methods
        # and check that they are valid
        model_names_default = [
            "sshleifer/distilbart-cnn-12-6",
            "distilbert-base-uncased-finetuned-sst-2-english",
            "dbmdz/bert-large-cased-finetuned-conll03-english",
        ]
        # no model names provided, set the default
        if not model_names:
            model_names = model_names_default
        # now assign model names for each of the methods
        # either to the provided model name or the default if one of the
        # task's models is set to None
        self.model_summary = (
            model_names[0] if model_names[0] else model_names_default[0]
        )
        self.model_sentiment = (
            model_names[1] if model_names[1] else model_names_default[1]
        )
        self.model_ner = model_names[2] if model_names[2] else model_names_default[2]

    def set_keys(self) -> dict:
        """Set the default keys for text analysis.

        Returns:
            dict: The dictionary with default text keys.
        """
        params = {"text": None, "text_language": None, "text_english": None}
        return params

    def _initialize_spacy(self):
        """Initialize the Spacy library for text analysis."""
        try:
            self.nlp = spacy.load("en_core_web_md")
        except Exception:
            spacy.cli.download("en_core_web_md")
            self.nlp = spacy.load("en_core_web_md")

    def analyse_image(self) -> dict:
        """Perform text extraction and analysis of the text.

        Returns:
            dict: The updated dictionary with text analysis results.
        """
        self.get_text_from_image()
        self.translate_text()
        self.remove_linebreaks()
        if self.analyse_text:
            self._run_spacy()
            self.clean_text()
            self.text_summary()
            self.text_sentiment_transformers()
            self.text_ner()
        return self.subdict

    def get_text_from_image(self):
        """Detect text on the image using Google Cloud Vision API."""
        path = self.subdict["filename"]
        try:
            client = vision.ImageAnnotatorClient()
        except DefaultCredentialsError:
            raise DefaultCredentialsError(
                "Please provide credentials for google cloud vision API, see https://cloud.google.com/docs/authentication/application-default-credentials."
            )
        with io.open(path, "rb") as image_file:
            content = image_file.read()
        image = vision.Image(content=content)
        # check for usual connection errors and retry if necessary
        try:
            response = client.text_detection(image=image)
        except grpc.RpcError as exc:
            print("Cloud vision API connection failed")
            print("Skipping this image ..{}".format(path))
            print("Connection failed with code {}: {}".format(exc.code(), exc))
        # here check if text was found on image
        if response:
            texts = response.text_annotations[0].description
            self.subdict["text"] = texts
        if response.error.message:
            print("Google Cloud Vision Error")
            raise ValueError(
                "{}\nFor more info on error messages, check: "
                "https://cloud.google.com/apis/design/errors".format(
                    response.error.message
                )
            )

    def translate_text(self):
        """Translate the detected text to English using the Translator object."""
        translated = self.translator.translate(self.subdict["text"])
        self.subdict["text_language"] = translated.src
        self.subdict["text_english"] = translated.text

    def remove_linebreaks(self):
        """Remove linebreaks from original and translated text."""
        if self.subdict["text"]:
            self.subdict["text"] = self.subdict["text"].replace("\n", " ")
            self.subdict["text_english"] = self.subdict["text_english"].replace(
                "\n", " "
            )

    def _run_spacy(self):
        """Generate Spacy doc object for further text analysis."""
        self.doc = self.nlp(self.subdict["text_english"])

    def clean_text(self):
        """Clean the text from unrecognized words and any numbers."""
        templist = []
        for token in self.doc:
            templist.append(
                token.text
            ) if token.pos_ != "NUM" and token.has_vector else None
        self.subdict["text_clean"] = " ".join(templist).rstrip().lstrip()

    def text_summary(self):
        """Generate a summary of the text using the Transformers pipeline."""
        # use the transformers pipeline to summarize the text
        # use the current default model - 03/2023
        max_number_of_characters = 3000
        pipe = pipeline(
            "summarization",
            model=self.model_summary,
            revision=self.revision_summary,
            min_length=5,
            max_length=20,
        )
        try:
            summary = pipe(self.subdict["text_english"][0:max_number_of_characters])
            self.subdict["text_summary"] = summary[0]["summary_text"]
        except IndexError:
            print(
                "Cannot provide summary for this object - please check that the text has been translated correctly."
            )
            print("Image: {}".format(self.subdict["filename"]))
            self.subdict["text_summary"] = None

    def text_sentiment_transformers(self):
        """Perform text classification for sentiment using the Transformers pipeline."""
        # use the transformers pipeline for text classification
        # use the current default model - 03/2023
        pipe = pipeline(
            "text-classification",
            model=self.model_sentiment,
            revision=self.revision_sentiment,
            truncation=True,
        )
        result = pipe(self.subdict["text_english"])
        self.subdict["sentiment"] = result[0]["label"]
        self.subdict["sentiment_score"] = round(result[0]["score"], 2)

    def text_ner(self):
        """Perform named entity recognition on the text using the Transformers pipeline."""
        # use the transformers pipeline for named entity recognition
        # use the current default model - 03/2023
        pipe = pipeline(
            "token-classification",
            model=self.model_ner,
            revision=self.revision_ner,
            aggregation_strategy="simple",
        )
        result = pipe(self.subdict["text_english"])
        self.subdict["entity"] = []
        self.subdict["entity_type"] = []
        for entity in result:
            self.subdict["entity"].append(entity["word"])
            self.subdict["entity_type"].append(entity["entity_group"])


class PostprocessText:
    def __init__(
        self,
        mydict: dict = None,
        use_csv: bool = False,
        csv_path: str = None,
        analyze_text: str = "text_english",
    ) -> None:
        """
        Initializes the PostprocessText class that handles the topic analysis.

        Args:
            mydict (dict, optional): Dictionary with textual data. Defaults to None.
            use_csv (bool, optional): Flag indicating whether to use a CSV file. Defaults to False.
            csv_path (str, optional): Path to the CSV file. Required if `use_csv` is True. Defaults to None.
            analyze_text (str, optional): Key for the text field to analyze. Defaults to "text_english".
        """
        self.use_csv = use_csv
        if mydict:
            print("Reading data from dict.")
            self.mydict = mydict
            self.list_text_english = self.get_text_dict(analyze_text)
        elif self.use_csv:
            print("Reading data from df.")
            self.df = pd.read_csv(csv_path, encoding="utf8")
            self.list_text_english = self.get_text_df(analyze_text)
        else:
            raise ValueError(
                "Please provide either dictionary with textual data or \
                              a csv file by setting `use_csv` to True and providing a \
                             `csv_path`."
            )
        # initialize spacy
        self._initialize_spacy()

    def _initialize_spacy(self):
        try:
            self.nlp = spacy.load(
                "en_core_web_md",
                exclude=["tagger", "parser", "ner", "attribute_ruler", "lemmatizer"],
            )
        except Exception:
            spacy.cli.download("en_core_web_md")
            self.nlp = spacy.load(
                "en_core_web_md",
                exclude=["tagger", "parser", "ner", "attribute_ruler", "lemmatizer"],
            )

    def analyse_topic(self, return_topics: int = 3) -> tuple:
        """
        Performs topic analysis using BERTopic.

        Args:
            return_topics (int, optional): Number of topics to return. Defaults to 3.

        Returns:
            tuple: A tuple containing the topic model, topic dataframe, and most frequent topics.
        """
        try:
            # unfortunately catching exceptions does not work here - need to figure out why
            self.topic_model = BERTopic(embedding_model=self.nlp)
        except TypeError:
            print("BERTopic excited with an error - maybe your dataset is too small?")
        self.topics, self.probs = self.topic_model.fit_transform(self.list_text_english)
        # return the topic list
        topic_df = self.topic_model.get_topic_info()
        # return the most frequent return_topics
        most_frequent_topics = []
        if len(topic_df) < return_topics:
            print("You requested more topics than are identified in your dataset -")
            print(
                "Returning only {} topics as these are all that have been found.".format(
                    len(topic_df)
                )
            )
        for i in range(min(return_topics, len(topic_df))):
            most_frequent_topics.append(self.topic_model.get_topic(i))
        return self.topic_model, topic_df, most_frequent_topics

    def get_text_dict(self, analyze_text: str) -> list:
        """
        Extracts text from the provided dictionary.

        Args:
            analyze_text (str): Key for the text field to analyze.

        Returns:
            list: A list of text extracted from the dictionary.
        """
        # use dict to put text_english or text_summary in list
        list_text_english = []
        for key in self.mydict.keys():
            if analyze_text not in self.mydict[key]:
                raise ValueError(
                    "Please check your provided dictionary - \
                no {} text data found.".format(
                        analyze_text
                    )
                )
            list_text_english.append(self.mydict[key][analyze_text])
        return list_text_english

    def get_text_df(self, analyze_text: str) -> list:
        """
        Extracts text from the provided dataframe.

        Args:
            analyze_text (str): Column name for the text field to analyze.

        Returns:
            list: A list of text extracted from the dataframe.
        """
        # use csv file to obtain dataframe and put text_english or text_summary in list
        # check that "text_english" or "text_summary" is there
        if analyze_text not in self.df:
            raise ValueError(
                "Please check your provided dataframe - \
                                no {} text data found.".format(
                    analyze_text
                )
            )
        return self.df[analyze_text].tolist()



from ammico.utils import AnalysisMethod
from ammico.objects_cvlib import ObjectCVLib
from ammico.objects_cvlib import init_default_objects


class ObjectDetectorClient(AnalysisMethod):
    def __init__(self):
        # The detector is set to CVLib by default
        self.detector = ObjectCVLib()

    def set_client_to_cvlib(self):
        """Set the object detection client to use CVLib."""
        self.detector = ObjectCVLib()

    def analyse_image(self, subdict=None):
        """Localize objects in the given image.

        Args:
            subdict (dict): The dictionary for an image expression instance.

        Returns:
            dict: The updated dictionary with object detection results.
        """
        return self.detector.analyse_image(subdict)


class ObjectDetector(AnalysisMethod):
    od_client = ObjectDetectorClient()

    def __init__(self, subdict: dict):
        super().__init__(subdict)
        self.subdict.update(self.set_keys())

    def set_keys(self):
        """Set the default object keys for analysis.

        Returns:
            dict: The dictionary with default object keys.
        """
        return init_default_objects()

    def analyse_image(self):
        """Perform object detection on the image.

        Returns:
            dict: The updated dictionary with object detection results.
        """
        self.subdict = ObjectDetector.od_client.analyse_image(self.subdict)
        return self.subdict

    @staticmethod
    def set_client_to_cvlib():
        """Set the object detection client to use CVLib."""
        ObjectDetector.od_client.set_client_to_cvlib()



from ammico.utils import AnalysisMethod
from torch import cuda, no_grad
from PIL import Image
from lavis.models import load_model_and_preprocess


class SummaryDetector(AnalysisMethod):
    def __init__(
        self,
        subdict: dict = {},
        summary_model_type: str = "base",
        analysis_type: str = "summary_and_questions",
        list_of_questions: str = None,
        summary_model=None,
        summary_vis_processors=None,
        summary_vqa_model=None,
        summary_vqa_vis_processors=None,
        summary_vqa_txt_processors=None,
    ) -> None:
        """
        SummaryDetector class for analysing images using the blip_caption model.

        Args:
            subdict (dict, optional): Dictionary containing the image to be analysed. Defaults to {}.
            summary_model_type (str, optional): Type of blip_caption model to use. Can be "base" or "large". Defaults to "base".
            analysis_type (str, optional): Type of analysis to perform. Can be "summary", "questions" or "summary_and_questions". Defaults to "summary_and_questions".
            list_of_questions (list, optional): List of questions to answer. Defaults to ["Are there people in the image?", "What is this picture about?"].
            summary_model ([type], optional): blip_caption model. Defaults to None.
            summary_vis_processors ([type], optional): Preprocessors for visual inputs. Defaults to None.
            summary_vqa_model ([type], optional): blip_vqa model. Defaults to None.
            summary_vqa_vis_processors ([type], optional): Preprocessors for vqa visual inputs. Defaults to None.
            summary_vqa_txt_processors ([type], optional): Preprocessors for vqa text inputs. Defaults to None.

        Raises:
            ValueError: If analysis_type is not one of "summary", "questions" or "summary_and_questions".

        Returns:
            None.
        """

        super().__init__(subdict)
        if analysis_type not in ["summary", "questions", "summary_and_questions"]:
            raise ValueError(
                "analysis_type must be one of 'summary', 'questions' or 'summary_and_questions'"
            )
        self.summary_device = "cuda" if cuda.is_available() else "cpu"
        allowed_model_types = ["base", "large"]
        if summary_model_type not in allowed_model_types:
            raise ValueError(
                "Model type is not allowed - please select one of {}".format(
                    allowed_model_types
                )
            )
        self.summary_model_type = summary_model_type
        self.analysis_type = analysis_type
        if list_of_questions is None:
            self.list_of_questions = [
                "Are there people in the image?",
                "What is this picture about?",
            ]
        elif (not isinstance(list_of_questions, list)) or (
            not all(isinstance(i, str) for i in list_of_questions)
        ):
            raise ValueError("list_of_questions must be a list of string (questions)")
        else:
            self.list_of_questions = list_of_questions
        if (
            (summary_model is None)
            and (summary_vis_processors is None)
            and (analysis_type != "questions")
        ):
            self.summary_model, self.summary_vis_processors = self.load_model(
                model_type=summary_model_type
            )
        else:
            self.summary_model = summary_model
            self.summary_vis_processors = summary_vis_processors
        if (
            (summary_vqa_model is None)
            and (summary_vqa_vis_processors is None)
            and (summary_vqa_txt_processors is None)
            and (analysis_type != "summary")
        ):
            (
                self.summary_vqa_model,
                self.summary_vqa_vis_processors,
                self.summary_vqa_txt_processors,
            ) = self.load_vqa_model()
        else:
            self.summary_vqa_model = summary_vqa_model
            self.summary_vqa_vis_processors = summary_vqa_vis_processors
            self.summary_vqa_txt_processors = summary_vqa_txt_processors

    def load_model_base(self):
        """
        Load base_coco blip_caption model and preprocessors for visual inputs from lavis.models.

        Args:

        Returns:
            model (torch.nn.Module): model.
            vis_processors (dict): preprocessors for visual inputs.
        """
        summary_model, summary_vis_processors, _ = load_model_and_preprocess(
            name="blip_caption",
            model_type="base_coco",
            is_eval=True,
            device=self.summary_device,
        )
        return summary_model, summary_vis_processors

    def load_model_large(self):
        """
        Load large_coco blip_caption model and preprocessors for visual inputs from lavis.models.

        Args:

        Returns:
            model (torch.nn.Module): model.
            vis_processors (dict): preprocessors for visual inputs.
        """
        summary_model, summary_vis_processors, _ = load_model_and_preprocess(
            name="blip_caption",
            model_type="large_coco",
            is_eval=True,
            device=self.summary_device,
        )
        return summary_model, summary_vis_processors

    def load_model(self, model_type: str):
        """
        Load blip_caption model and preprocessors for visual inputs from lavis.models.

        Args:
            model_type (str): type of the model.

        Returns:
            model (torch.nn.Module): model.
            vis_processors (dict): preprocessors for visual inputs.
        """
        select_model = {
            "base": SummaryDetector.load_model_base,
            "large": SummaryDetector.load_model_large,
        }
        summary_model, summary_vis_processors = select_model[model_type](self)
        return summary_model, summary_vis_processors

    def load_vqa_model(self):
        """
        Load blip_vqa model and preprocessors for visual and text inputs from lavis.models.

        Args:

        Returns:
            model (torch.nn.Module): model.
            vis_processors (dict): preprocessors for visual inputs.
            txt_processors (dict): preprocessors for text inputs.

        """
        (
            summary_vqa_model,
            summary_vqa_vis_processors,
            summary_vqa_txt_processors,
        ) = load_model_and_preprocess(
            name="blip_vqa",
            model_type="vqav2",
            is_eval=True,
            device=self.summary_device,
        )
        return summary_vqa_model, summary_vqa_vis_processors, summary_vqa_txt_processors

    def analyse_image(self):
        """
        Analyse image with blip_caption model.

        Args:

        Returns:
            self.subdict (dict): dictionary with analysis results.
        """
        if self.analysis_type == "summary_and_questions":
            self.analyse_summary()
            self.analyse_questions(self.list_of_questions)
        elif self.analysis_type == "summary":
            self.analyse_summary()
        elif self.analysis_type == "questions":
            self.analyse_questions(self.list_of_questions)

        return self.subdict

    def analyse_summary(self):
        """
        Create 1 constant and 3 non deterministic captions for image.

        Args:

        Returns:
            self.subdict (dict): dictionary with analysis results.
        """

        path = self.subdict["filename"]
        raw_image = Image.open(path).convert("RGB")
        image = (
            self.summary_vis_processors["eval"](raw_image)
            .unsqueeze(0)
            .to(self.summary_device)
        )
        with no_grad():
            self.subdict["const_image_summary"] = self.summary_model.generate(
                {"image": image}
            )[0]
            self.subdict["3_non-deterministic summary"] = self.summary_model.generate(
                {"image": image}, use_nucleus_sampling=True, num_captions=3
            )
        return self.subdict

    def analyse_questions(self, list_of_questions: list[str]) -> dict:
        """
        Generate answers to free-form questions about image written in natural language.

        Args:
            list_of_questions (list[str]): list of questions.

        Returns:
            self.subdict (dict): dictionary with answers to questions.
        """
        if (
            (self.summary_vqa_model is None)
            and (self.summary_vqa_vis_processors is None)
            and (self.summary_vqa_txt_processors is None)
        ):
            (
                self.summary_vqa_model,
                self.summary_vqa_vis_processors,
                self.summary_vqa_txt_processors,
            ) = load_model_and_preprocess(
                name="blip_vqa",
                model_type="vqav2",
                is_eval=True,
                device=self.summary_device,
            )
        if len(list_of_questions) > 0:
            path = self.subdict["filename"]
            raw_image = Image.open(path).convert("RGB")
            image = (
                self.summary_vqa_vis_processors["eval"](raw_image)
                .unsqueeze(0)
                .to(self.summary_device)
            )
            question_batch = []
            for quest in list_of_questions:
                question_batch.append(self.summary_vqa_txt_processors["eval"](quest))
            batch_size = len(list_of_questions)
            image_batch = image.repeat(batch_size, 1, 1, 1)

            with no_grad():
                answers_batch = self.summary_vqa_model.predict_answers(
                    samples={"image": image_batch, "text_input": question_batch},
                    inference_method="generate",
                )

            for q, a in zip(list_of_questions, answers_batch):
                self.subdict[q] = a

        else:
            print("Please, enter list of questions")
        return self.subdict



import cv2
import numpy as np
import os
import shutil
import pathlib
from tensorflow.keras.models import load_model
from tensorflow.keras.applications.mobilenet_v2 import preprocess_input
from tensorflow.keras.preprocessing.image import img_to_array
from deepface import DeepFace
from retinaface import RetinaFace
from ammico.utils import DownloadResource, AnalysisMethod


DEEPFACE_PATH = ".deepface"


def deepface_symlink_processor(name):
    def _processor(fname, action, pooch):
        if not os.path.exists(name):
            # symlink does not work on windows
            # use copy if running on windows
            if os.name != "nt":
                os.symlink(fname, name)
            else:
                shutil.copy(fname, name)
        return fname

    return _processor


face_mask_model = DownloadResource(
    url="https://github.com/chandrikadeb7/Face-Mask-Detection/raw/v1.0.0/mask_detector.model",
    known_hash="sha256:d0b30e2c7f8f187c143d655dee8697fcfbe8678889565670cd7314fb064eadc8",
)

deepface_age_model = DownloadResource(
    url="https://github.com/serengil/deepface_models/releases/download/v1.0/age_model_weights.h5",
    known_hash="sha256:0aeff75734bfe794113756d2bfd0ac823d51e9422c8961125b570871d3c2b114",
    processor=deepface_symlink_processor(
        pathlib.Path.home().joinpath(DEEPFACE_PATH, "weights", "age_model_weights.h5")
    ),
)

deepface_face_expression_model = DownloadResource(
    url="https://github.com/serengil/deepface_models/releases/download/v1.0/facial_expression_model_weights.h5",
    known_hash="sha256:e8e8851d3fa05c001b1c27fd8841dfe08d7f82bb786a53ad8776725b7a1e824c",
    processor=deepface_symlink_processor(
        pathlib.Path.home().joinpath(
            ".deepface", "weights", "facial_expression_model_weights.h5"
        )
    ),
)

deepface_gender_model = DownloadResource(
    url="https://github.com/serengil/deepface_models/releases/download/v1.0/gender_model_weights.h5",
    known_hash="sha256:45513ce5678549112d25ab85b1926fb65986507d49c674a3d04b2ba70dba2eb5",
    processor=deepface_symlink_processor(
        pathlib.Path.home().joinpath(
            DEEPFACE_PATH, "weights", "gender_model_weights.h5"
        )
    ),
)

deepface_race_model = DownloadResource(
    url="https://github.com/serengil/deepface_models/releases/download/v1.0/race_model_single_batch.h5",
    known_hash="sha256:eb22b28b1f6dfce65b64040af4e86003a5edccb169a1a338470dde270b6f5e54",
    processor=deepface_symlink_processor(
        pathlib.Path.home().joinpath(
            DEEPFACE_PATH, "weights", "race_model_single_batch.h5"
        )
    ),
)

retinaface_model = DownloadResource(
    url="https://github.com/serengil/deepface_models/releases/download/v1.0/retinaface.h5",
    known_hash="sha256:ecb2393a89da3dd3d6796ad86660e298f62a0c8ae7578d92eb6af14e0bb93adf",
    processor=deepface_symlink_processor(
        pathlib.Path.home().joinpath(DEEPFACE_PATH, "weights", "retinaface.h5")
    ),
)


class EmotionDetector(AnalysisMethod):
    def __init__(
        self,
        subdict: dict,
        emotion_threshold: float = 50.0,
        race_threshold: float = 50.0,
    ) -> None:
        """
        Initializes the EmotionDetector object.

        Args:
            subdict (dict): The dictionary to store the analysis results.
            emotion_threshold (float): The threshold for detecting emotions (default: 50.0).
            race_threshold (float): The threshold for detecting race (default: 50.0).
        """
        super().__init__(subdict)
        self.subdict.update(self.set_keys())
        self.emotion_threshold = emotion_threshold
        self.race_threshold = race_threshold
        self.emotion_categories = {
            "angry": "Negative",
            "disgust": "Negative",
            "fear": "Negative",
            "sad": "Negative",
            "happy": "Positive",
            "surprise": "Neutral",
            "neutral": "Neutral",
        }

    def set_keys(self) -> dict:
        """
        Sets the initial parameters for the analysis.

        Returns:
            dict: The dictionary with initial parameter values.
        """
        params = {
            "face": "No",
            "multiple_faces": "No",
            "no_faces": 0,
            "wears_mask": ["No"],
            "age": [None],
            "gender": [None],
            "race": [None],
            "emotion": [None],
            "emotion (category)": [None],
        }
        return params

    def analyse_image(self) -> dict:
        """
        Performs facial expression analysis on the image.

        Returns:
            dict: The updated subdict dictionary with analysis results.
        """
        return self.facial_expression_analysis()

    def analyze_single_face(self, face: np.ndarray) -> dict:
        """
        Analyzes the features of a single face.

        Args:
            face (np.ndarray): The face image array.

        Returns:
            dict: The analysis results for the face.
        """
        fresult = {}
        # Determine whether the face wears a mask
        fresult["wears_mask"] = self.wears_mask(face)
        # Adapt the features we are looking for depending on whether a mask is worn.
        # White masks screw race detection, emotion detection is useless.
        actions = ["age", "gender"]
        if not fresult["wears_mask"]:
            actions = actions + ["race", "emotion"]
        # Ensure that all data has been fetched by pooch
        deepface_age_model.get()
        deepface_face_expression_model.get()
        deepface_gender_model.get()
        deepface_race_model.get()
        # Run the full DeepFace analysis
        fresult.update(
            DeepFace.analyze(
                img_path=face,
                actions=actions,
                prog_bar=False,
                detector_backend="skip",
            )
        )
        # We remove the region, as the data is not correct - after all we are
        # running the analysis on a subimage.
        del fresult["region"]
        return fresult

    def facial_expression_analysis(self) -> dict:
        """
        Performs facial expression analysis on the image.

        Returns:
            dict: The updated subdict dictionary with analysis results.
        """
        # Find (multiple) faces in the image and cut them
        retinaface_model.get()

        faces = RetinaFace.extract_faces(self.subdict["filename"])
        # If no faces are found, we return empty keys
        if len(faces) == 0:
            return self.subdict
        # Sort the faces by sight to prioritize prominent faces
        faces = list(reversed(sorted(faces, key=lambda f: f.shape[0] * f.shape[1])))
        self.subdict["face"] = "Yes"
        self.subdict["multiple_faces"] = "Yes" if len(faces) > 1 else "No"
        self.subdict["no_faces"] = len(faces) if len(faces) <= 15 else 99
        # note number of faces being identified
        result = {"number_faces": len(faces) if len(faces) <= 3 else 3}
        # We limit ourselves to three faces
        for i, face in enumerate(faces[:3]):
            result[f"person{ i+1 }"] = self.analyze_single_face(face)
        self.clean_subdict(result)
        return self.subdict

    def clean_subdict(self, result: dict) -> dict:
        """
        Cleans the subdict dictionary by converting results into appropriate formats.

        Args:
            result (dict): The analysis results.
        Returns:
            dict: The updated subdict dictionary.
        """
        # Each person subdict converted into list for keys
        self.subdict["wears_mask"] = []
        self.subdict["age"] = []
        self.subdict["gender"] = []
        self.subdict["race"] = []
        self.subdict["emotion"] = []
        self.subdict["emotion (category)"] = []
        for i in range(result["number_faces"]):
            person = "person{}".format(i + 1)
            self.subdict["wears_mask"].append(
                "Yes" if result[person]["wears_mask"] else "No"
            )
            self.subdict["age"].append(result[person]["age"])
            # Gender is now reported as a list of dictionaries.
            # Each dict represents one face.
            # Each dict contains probability for Woman and Man.
            # We take only the higher probability result for each dict.
            self.subdict["gender"].append(result[person]["gender"])
            # Race and emotion are only detected if a person does not wear a mask
            if result[person]["wears_mask"]:
                self.subdict["race"].append(None)
                self.subdict["emotion"].append(None)
                self.subdict["emotion (category)"].append(None)
            elif not result[person]["wears_mask"]:
                # Check whether the race threshold was exceeded
                if (
                    result[person]["race"][result[person]["dominant_race"]]
                    > self.race_threshold
                ):
                    self.subdict["race"].append(result[person]["dominant_race"])
                else:
                    self.subdict["race"].append(None)

                # Check whether the emotion threshold was exceeded
                if (
                    result[person]["emotion"][result[person]["dominant_emotion"]]
                    > self.emotion_threshold
                ):
                    self.subdict["emotion"].append(result[person]["dominant_emotion"])
                    self.subdict["emotion (category)"].append(
                        self.emotion_categories[result[person]["dominant_emotion"]]
                    )
                else:
                    self.subdict["emotion"].append(None)
                    self.subdict["emotion (category)"].append(None)
        return self.subdict

    def wears_mask(self, face: np.ndarray) -> bool:
        """
        Determines whether a face wears a mask.

        Args:
            face (np.ndarray): The face image array.

        Returns:
            bool: True if the face wears a mask, False otherwise.
        """
        global mask_detection_model
        # Preprocess the face to match the assumptions of the face mask detection model
        face = cv2.cvtColor(face, cv2.COLOR_BGR2RGB)
        face = cv2.resize(face, (224, 224))
        face = img_to_array(face)
        face = preprocess_input(face)
        face = np.expand_dims(face, axis=0)
        # Lazily load the model
        mask_detection_model = load_model(face_mask_model.get())
        # Run the model
        mask, without_mask = mask_detection_model.predict(face)[0]
        # Convert from np.bool_ to bool to later be able to serialize the result
        return bool(mask > without_mask)



import os
import ntpath
import cv2
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from typing import List, Optional, Tuple, Union


MIN_MATCH_COUNT = 6
FLANN_INDEX_KDTREE = 1


# use this function to visualize the matches from sift
def draw_matches(
    matches: List,
    img1: np.ndarray,
    img2: np.ndarray,
    kp1: List[cv2.KeyPoint],
    kp2: List[cv2.KeyPoint],
) -> None:
    """Visualize the matches from SIFT.

    Args:
        matches (list[cv2.Match]): List of cv2.Match matches on the image.
        img1 (np.ndarray): The reference image.
        img2 (np.ndarray): The social media post.
        kp1 (list[cv2.KeyPoint]): List of keypoints from the first image.
        kp2 (list[cv2.KeyPoint]): List of keypoints from the second image.
    """
    if len(matches) > MIN_MATCH_COUNT:
        # Estimate homography between template and scene
        src_pts = np.float32([kp1[m.queryIdx].pt for m in matches]).reshape(-1, 1, 2)
        dst_pts = np.float32([kp2[m.trainIdx].pt for m in matches]).reshape(-1, 1, 2)
        M = cv2.findHomography(src_pts, dst_pts, cv2.RANSAC, 5.0)[0]
        if not isinstance(M, np.ndarray):
            print("Could not match images for drawing.")
            return
        # Draw detected template in scene image
        h = img1.shape[0]
        w = img1.shape[1]
        pts = np.float32([[0, 0], [0, h - 1], [w - 1, h - 1], [w - 1, 0]]).reshape(
            -1, 1, 2
        )
        dst = cv2.perspectiveTransform(pts, M)
        img2 = cv2.polylines(img2, [np.int32(dst)], True, 255, 3, cv2.LINE_AA)
        h1 = img1.shape[0]
        h2 = img2.shape[0]
        w1 = img1.shape[1]
        w2 = img2.shape[1]
        nwidth = w1 + w2
        nheight = max(h1, h2)
        hdif = int((h2 - h1) / 2)
        newimg = np.zeros((nheight, nwidth, 3), np.uint8)
        for i in range(3):
            newimg[hdif : hdif + h1, :w1, i] = img1
            newimg[:h2, w1 : w1 + w2, i] = img2
        # Draw SIFT keypoint matches
        for m in matches:
            pt1 = (int(kp1[m.queryIdx].pt[0]), int(kp1[m.queryIdx].pt[1] + hdif))
            pt2 = (int(kp2[m.trainIdx].pt[0] + w1), int(kp2[m.trainIdx].pt[1]))
            cv2.line(newimg, pt1, pt2, (255, 0, 0))
        plt.imshow(newimg)
        plt.show()
    else:
        print("Not enough matches are found - %d/%d" % (len(matches), MIN_MATCH_COUNT))


def matching_points(
    img1: np.ndarray, img2: np.ndarray
) -> Tuple[cv2.DMatch, List[cv2.KeyPoint], List[cv2.KeyPoint]]:
    """Computes keypoint matches using the SIFT algorithm between two images.

    Args:
        img1 (np.ndarray): The reference image.
        img2 (np.ndarray): The social media post.
    Returns:
        cv2.DMatch: List of filtered keypoint matches.
        cv2.KeyPoint: List of keypoints from the first image.
        cv2.KeyPoint: List of keypoints from the second image.
    """
    img1 = cv2.cvtColor(img1, cv2.COLOR_BGR2GRAY)
    img2 = cv2.cvtColor(img2, cv2.COLOR_BGR2GRAY)
    sift = cv2.SIFT_create()
    kp1, des1 = sift.detectAndCompute(img1, None)
    kp2, des2 = sift.detectAndCompute(img2, None)

    # Convert descriptors to float32
    des1 = np.float32(des1)
    des2 = np.float32(des2)
    # Initialize and use FLANN
    index_params = dict(algorithm=FLANN_INDEX_KDTREE, trees=5)
    search_params = dict(checks=50)
    flann = cv2.FlannBasedMatcher(index_params, search_params)
    matches = flann.knnMatch(des1, des2, k=2)
    filtered_matches = []
    for m, n in matches:
        # Apply ratio test to filter out ambiguous matches
        if m.distance < 0.7 * n.distance:
            filtered_matches.append(m)
    return filtered_matches, kp1, kp2


def kp_from_matches(matches, kp1: np.ndarray, kp2: np.ndarray) -> Tuple[Tuple, Tuple]:
    """Extract the match indices from the keypoints.

    Args:
        kp1 (np.ndarray): Key points of the matches,
        kp2 (np.ndarray):  Key points of the matches,
    Returns:
        tuple: Index of the descriptor in the list of train descriptors.
        tuple: index of the descriptor in the list of query descriptors.
    """
    kp1 = np.float32([kp1[m.queryIdx].pt for m in matches])
    kp2 = np.float32([kp2[m.trainIdx].pt for m in matches])
    return kp1, kp2


def compute_crop_corner(
    matches: cv2.DMatch,
    kp1: np.ndarray,
    kp2: np.ndarray,
    region: int = 30,
    h_margin: int = 0,
    v_margin: int = 5,
    min_match: int = 6,
) -> Optional[Tuple[int, int]]:
    """Estimate the position on the image from where to crop.

    Args:
        matches (cv2.DMatch): The matched objects on the image.
        kp1 (np.ndarray): Key points of the matches for the reference image.
        kp2 (np.ndarray): Key points of the matches for the social media posts.
        region (int, optional): Area to consider around the keypoints.
            Defaults to 30.
        h_margin (int, optional): Horizontal margin to subtract from the minimum
            horizontal position. Defaults to 0.
        v_margin (int, optional): Vertical margin to subtract from the minimum
            vertical position. Defaults to 5.
        min_match: Minimum number of matches required. Defaults to 6.
    Returns:
        tuple, optional: Tuple of vertical and horizontal crop corner coordinates.
    """
    kp1, kp2 = kp_from_matches(matches, kp1, kp2)
    ys = kp2[:, 1]
    covers = []

    # Compute the number of keypoints within the region around each y-coordinate
    for y in ys:
        ys_c = ys - y
        series = pd.Series(ys_c)
        is_between = series.between(0, region)
        covers.append(is_between.sum())
    covers = np.array(covers)
    if covers.max() < min_match:
        return None
    kp_id = ys[covers.argmax()]
    v = int(kp_id) - v_margin if int(kp_id) > v_margin else int(kp_id)

    hs = []

    # Find the minimum x-coordinate within the region around the selected y-coordinate
    for kp in kp2:
        if 0 <= kp[1] - v <= region:
            hs.append(kp[0])
    # do not use margin if h < image width/2, else use margin
    h = int(np.min(hs)) - h_margin if int(np.min(hs)) > h_margin else 0
    return v, h


def crop_posts_image(
    ref_view: List,
    view: np.ndarray,
) -> Union[None, Tuple[np.ndarray, int, int, int]]:
    """Crop the social media post to exclude additional comments. Sometimes also crops the
    image part of the post - this is put back in later.

    Args:
        ref_views (list): List of all the reference images (as numpy arrays) that signify
            below which regions should be cropped.
        view (np.ndarray): The image to crop.
    Returns:
        np.ndarray: The cropped social media post.
    """
    filtered_matches, kp1, kp2 = matching_points(ref_view, view)
    if len(filtered_matches) < MIN_MATCH_COUNT:
        # not enough matches found
        # print("Found too few matches - {}".format(filtered_matches))
        return None
    corner = compute_crop_corner(filtered_matches, kp1, kp2)
    if corner is None:
        # no cropping corner found
        # print("Found no corner")
        return None
    v, h = corner
    # if the match is on the right-hand side of the image,
    # it is likely that there is an image to the left
    # that should not be cropped
    # in this case, we adjust the margin for the text to be
    # cropped to `correct_margin`
    # if the match is more to the left on the image, we assume
    # it starts at horizontal position 0 and do not want to
    # cut off any characters from the text
    correct_margin = 30
    if h >= view.shape[1] / 2:
        h = h - correct_margin
    else:
        h = 0
    crop_view = view[0:v, h:, :]
    return crop_view, len(filtered_matches), v, h


def crop_posts_from_refs(
    ref_views: List,
    view: np.ndarray,
    plt_match: bool = False,
    plt_crop: bool = False,
    plt_image: bool = False,
) -> np.ndarray:
    """Crop the social media post comments from the image.

    Args:
        ref_views (list): List of all the reference images (as numpy arrays) that signify
            below which regions should be cropped.
        view (np.ndarray): The image to crop.
    Returns:
        np.ndarray: The cropped social media post.
    """
    crop_view = None
    # initialize the number of found matches per reference to zero
    # so later we can select the reference with the most matches
    max_matchs = 0
    rte = None
    found_match = False
    for ref_view in ref_views:
        rte = crop_posts_image(ref_view, view)
        if rte is not None:
            crop_img, match_num, v, h = rte
            if match_num > max_matchs:
                # find the reference with the most matches to crop accordingly
                crop_view = crop_img
                final_ref = ref_view
                final_v = v
                final_h = h
                max_matchs = match_num
                found_match = True

    if found_match and plt_match:
        # plot the match
        filtered_matches, kp1, kp2 = matching_points(final_ref, view)
        img1 = cv2.cvtColor(final_ref, cv2.COLOR_BGR2GRAY)
        img2 = cv2.cvtColor(view, cv2.COLOR_BGR2GRAY)
        draw_matches(filtered_matches, img1, img2, kp1, kp2)

    if found_match and plt_crop:
        # plot the cropped image
        view2 = view.copy()
        view2[final_v, :, 0:3] = [255, 0, 0]
        view2[:, final_h, 0:3] = [255, 0, 0]
        plt.imshow(cv2.cvtColor(view2, cv2.COLOR_BGR2RGB))
        plt.show()
        plt.imshow(cv2.cvtColor(crop_view, cv2.COLOR_BGR2RGB))
        plt.show()

    if found_match and final_h >= view.shape[1] / 2:
        # here it would crop the actual image from the social media post
        # to avoid this, we check the position from where it would crop
        # if > than half of the width of the image, also keep all that is
        # on the left-hand side of the crop
        crop_post = crop_image_from_post(view, final_h)
        if plt_image:
            # plot the image part of the social media post
            plt.imshow(cv2.cvtColor(crop_post, cv2.COLOR_BGR2RGB))
            plt.show()
        # now concatenate the image and the text part
        crop_view = paste_image_and_comment(crop_post, crop_view)
    return crop_view


def crop_image_from_post(view: np.ndarray, final_h: int) -> np.ndarray:
    """Crop the image part from the social media post.

    Args:
        view (np.ndarray): The image to be cropped.
        final_h: The horizontal position up to which should be cropped.
    Returns:
        np.ndarray: The cropped image part."""
    crop_post = view[:, 0:final_h, :]
    return crop_post


def paste_image_and_comment(crop_post: np.ndarray, crop_view: np.ndarray) -> np.ndarray:
    """Paste the image part and the text part together without the unecessary comments.

    Args:
        crop_post (np.ndarray): The cropped image part of the social media post.
        crop_view (np.ndarray): The cropped text part of the social media post.
    Returns:
        np.ndarray: The image and text part of the social media post in one image."""
    h1, w1 = crop_post.shape[:2]
    h2, w2 = crop_view.shape[:2]
    image_all = np.zeros((max(h1, h2), w1 + w2, 3), np.uint8)
    image_all[:h1, :w1, :3] = crop_post
    image_all[:h2, w1 : w1 + w2, :3] = crop_view
    return image_all


def crop_media_posts(
    files, ref_files, save_crop_dir, plt_match=False, plt_crop=False, plt_image=False
) -> None:
    """Crop social media posts so that comments beyond the first comment/post are cut off.

    Args:
        files (list): List of all the files to be cropped.
        ref_files (list): List of all the reference images that signify
            below which regions should be cropped.
        save_crop_dir (str): Directory where to write the cropped social media posts to.
        plt_match (Bool, optional): Display the matched areas on the social media post.
            Defaults to False.
        plt_crop (Bool, optional): Display the cropped text part of the social media post.
            Defaults to False.
        plt_image (Bool, optional): Display the image part of the social media post.
            Defaults to False.
    """

    # get the reference images with regions that signify areas to crop
    ref_views = []
    for ref_file in ref_files:
        ref_view = cv2.imread(ref_file)
        ref_views.append(ref_view)
    # parse through the social media posts to be cropped
    for crop_file in files:
        view = cv2.imread(crop_file)
        print("Doing file {}".format(crop_file))
        crop_view = crop_posts_from_refs(
            ref_views,
            view,
            plt_match=plt_match,
            plt_crop=plt_crop,
            plt_image=plt_image,
        )
        if crop_view is not None:
            # save the image to the provided folder
            filename = ntpath.basename(crop_file)
            save_path = os.path.join(save_crop_dir, filename)
            save_path = save_path.replace("\\", "/")
            cv2.imwrite(save_path, crop_view)



from ammico.utils import AnalysisMethod
import torch
import torch.nn.functional as Func
import requests
import lavis
import os
import numpy as np
from PIL import Image
from skimage import transform as skimage_transform
from scipy.ndimage import filters
from matplotlib import pyplot as plt
from IPython.display import display
from lavis.models import load_model_and_preprocess, load_model, BlipBase
from lavis.processors import load_processor


class MultimodalSearch(AnalysisMethod):
    def __init__(self, subdict: dict) -> None:
        super().__init__(subdict)

    multimodal_device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    def load_feature_extractor_model_blip2(self, device: str = "cpu"):
        """
        Load pretrain blip2_feature_extractor model and preprocessors for visual and text inputs from lavis.models.

        Args:
            device (str): device to use. Can be "cpu" or "cuda". Default: "cpu".

        Returns:
            model (torch.nn.Module): model.
            vis_processors (dict): preprocessors for visual inputs.
            txt_processors (dict): preprocessors for text inputs.
        """
        model, vis_processors, txt_processors = load_model_and_preprocess(
            name="blip2_feature_extractor",
            model_type="pretrain",
            is_eval=True,
            device=device,
        )
        return model, vis_processors, txt_processors

    def load_feature_extractor_model_blip(self, device: str = "cpu"):
        """
        Load base blip_feature_extractor model and preprocessors for visual and text inputs from lavis.models.

        Args:
            device (str): device to use. Can be "cpu" or "cuda". Default: "cpu".

        Returns:
            model (torch.nn.Module): model.
            vis_processors (dict): preprocessors for visual inputs.
            txt_processors (dict): preprocessors for text inputs.
        """
        model, vis_processors, txt_processors = load_model_and_preprocess(
            name="blip_feature_extractor",
            model_type="base",
            is_eval=True,
            device=device,
        )
        return model, vis_processors, txt_processors

    def load_feature_extractor_model_albef(self, device: str = "cpu"):
        """
        Load base albef_feature_extractor model and preprocessors for visual and text inputs from lavis.models.

        Args:
            device (str): device to use. Can be "cpu" or "cuda". Default: "cpu".

        Returns:
            model (torch.nn.Module): model.
            vis_processors (dict): preprocessors for visual inputs.
            txt_processors (dict): preprocessors for text inputs.
        """
        model, vis_processors, txt_processors = load_model_and_preprocess(
            name="albef_feature_extractor",
            model_type="base",
            is_eval=True,
            device=device,
        )
        return model, vis_processors, txt_processors

    def load_feature_extractor_model_clip_base(self, device: str = "cpu"):
        """
        Load base clip_feature_extractor model and preprocessors for visual and text inputs from lavis.models.

        Args:
            device (str): device to use. Can be "cpu" or "cuda". Default: "cpu".

        Returns:
            model (torch.nn.Module): model.
            vis_processors (dict): preprocessors for visual inputs.
            txt_processors (dict): preprocessors for text inputs.
        """
        model, vis_processors, txt_processors = load_model_and_preprocess(
            name="clip_feature_extractor",
            model_type="base",
            is_eval=True,
            device=device,
        )
        return model, vis_processors, txt_processors

    def load_feature_extractor_model_clip_vitl14(self, device: str = "cpu"):
        """
        Load ViT-L-14 clip_feature_extractor model and preprocessors for visual and text inputs from lavis.models.

        Args:
            device (str): device to use. Can be "cpu" or "cuda". Default: "cpu".

        Returns:
            model (torch.nn.Module): model.
            vis_processors (dict): preprocessors for visual inputs.
            txt_processors (dict): preprocessors for text inputs.
        """
        model, vis_processors, txt_processors = load_model_and_preprocess(
            name="clip_feature_extractor",
            model_type="ViT-L-14",
            is_eval=True,
            device=device,
        )
        return model, vis_processors, txt_processors

    def load_feature_extractor_model_clip_vitl14_336(self, device: str = "cpu"):
        """
        Load ViT-L-14-336 clip_feature_extractor model and preprocessors for visual and text inputs from lavis.models.

        Args:
            device (str): device to use. Can be "cpu" or "cuda". Default: "cpu".

        Returns:
            model (torch.nn.Module): model.
            vis_processors (dict): preprocessors for visual inputs.
            txt_processors (dict): preprocessors for text inputs.
        """
        model, vis_processors, txt_processors = load_model_and_preprocess(
            name="clip_feature_extractor",
            model_type="ViT-L-14-336",
            is_eval=True,
            device=device,
        )
        return model, vis_processors, txt_processors

    def read_img(self, filepath: str) -> Image:
        """
        Load Image from filepath.

        Args:
            filepath (str): path to image.

        Returns:
            raw_image (PIL.Image): image.
        """
        raw_image = Image.open(filepath).convert("RGB")
        return raw_image

    def read_and_process_images(self, image_paths: list, vis_processor) -> tuple:
        """
        Read and process images with vis_processor.

        Args:
            image_paths (str): paths to images.
            vis_processor (dict): preprocessors for visual inputs.

        Returns:
            raw_images (list): list of images.
            images_tensors (torch.Tensor): tensors of images stacked in device.
        """
        raw_images = [MultimodalSearch.read_img(self, path) for path in image_paths]
        images = [
            vis_processor["eval"](r_img)
            .unsqueeze(0)
            .to(MultimodalSearch.multimodal_device)
            for r_img in raw_images
        ]
        images_tensors = torch.stack(images)

        return raw_images, images_tensors

    def extract_image_features_blip2(
        self, model, images_tensors: torch.Tensor
    ) -> torch.Tensor:
        """
        Extract image features from images_tensors with blip2_feature_extractor model.

        Args:
            model (torch.nn.Module): model.
            images_tensors (torch.Tensor): tensors of images stacked in device.

        Returns:
            features_image_stacked (torch.Tensor): tensors of images features stacked in device.
        """
        with torch.cuda.amp.autocast(
            enabled=(MultimodalSearch.multimodal_device != torch.device("cpu"))
        ):
            features_image = [
                model.extract_features({"image": ten, "text_input": ""}, mode="image")
                for ten in images_tensors
            ]
            features_image_stacked = torch.stack(
                [feat.image_embeds_proj[:, 0, :].squeeze(0) for feat in features_image]
            )
        return features_image_stacked

    def extract_image_features_clip(
        self, model, images_tensors: torch.Tensor
    ) -> torch.Tensor:
        """
        Extract image features from images_tensors with clip_feature_extractor model.

        Args:
            model (torch.nn.Module): model.
            images_tensors (torch.Tensor): tensors of images stacked in device.

        Returns:
            features_image_stacked (torch.Tensor): tensors of images features stacked in device.
        """
        features_image = [
            model.extract_features({"image": ten}) for ten in images_tensors
        ]
        features_image_stacked = torch.stack(
            [Func.normalize(feat.float(), dim=-1).squeeze(0) for feat in features_image]
        )
        return features_image_stacked

    def extract_image_features_basic(
        self, model, images_tensors: torch.Tensor
    ) -> torch.Tensor:
        """
        Extract image features from images_tensors with blip_feature_extractor or albef_feature_extractor model.

        Args:
            model (torch.nn.Module): model.
            images_tensors (torch.Tensor): tensors of images stacked in device.

        Returns:
            features_image_stacked (torch.Tensor): tensors of images features stacked in device.
        """
        features_image = [
            model.extract_features({"image": ten, "text_input": ""}, mode="image")
            for ten in images_tensors
        ]
        features_image_stacked = torch.stack(
            [feat.image_embeds_proj[:, 0, :].squeeze(0) for feat in features_image]
        )
        return features_image_stacked

    def save_tensors(
        self,
        model_type: str,
        features_image_stacked: torch.Tensor,
        name: str = "saved_features_image.pt",
        path: str = "./saved_tensors/",
    ) -> str:
        """
        Save tensors as binary to given path.

        Args:
            model_type (str): type of the model.
            features_image_stacked (torch.Tensor): tensors of images features stacked in device.
            name (str): name of the file. Default: "saved_features_image.pt".
            path (str): path to save the file. Default: "./saved_tensors/".

        Returns:
            name (str): name of the file.
        """
        if not os.path.exists(path):
            os.makedirs(path)
        with open(
            str(path)
            + str(len(features_image_stacked))
            + "_"
            + model_type
            + "_"
            + name,
            "wb",
        ) as f:
            torch.save(features_image_stacked, f)
        return name

    def load_tensors(self, name: str) -> torch.Tensor:
        """
        Load tensors from given path.

        Args:
            name (str): name of the file.

        Returns:
            features_image_stacked (torch.Tensor): tensors of images features.
        """
        features_image_stacked = torch.load(name)
        return features_image_stacked

    def extract_text_features(self, model, text_input: str) -> torch.Tensor:
        """
        Extract text features from text_input with feature_extractor model.

        Args:
            model (torch.nn.Module): model.
            text_input (str): text.

        Returns:
            features_text (torch.Tensor): tensors of text features.
        """
        sample_text = {"text_input": [text_input]}
        features_text = model.extract_features(sample_text, mode="text")

        return features_text

    def parsing_images(
        self,
        model_type: str,
        path_to_save_tensors: str = "./saved_tensors/",
        path_to_load_tensors: str = None,
    ) -> tuple:
        """
        Parsing images with feature_extractor model.

        Args:
            model_type (str): type of the model.
            path_to_save_tensors (str): path to save the tensors. Default: "./saved_tensors/".
            path_to_load_tensors (str): path to load the tesors. Default: None.

        Returns:
            model (torch.nn.Module): model.
            vis_processors (dict): preprocessors for visual inputs.
            txt_processors (dict): preprocessors for text inputs.
            image_keys (list): sorted list of image keys.
            image_names (list): sorted list of image names.
            features_image_stacked (torch.Tensor): tensors of images features stacked in device.
        """
        if model_type in ("clip_base", "clip_vitl14_336", "clip_vitl14"):
            path_to_lib = lavis.__file__[:-11] + "models/clip_models/"
            url = "https://raw.githubusercontent.com/salesforce/LAVIS/main/lavis/models/clip_models/bpe_simple_vocab_16e6.txt.gz"
            r = requests.get(url, allow_redirects=False)
            open(path_to_lib + "bpe_simple_vocab_16e6.txt.gz", "wb").write(r.content)

        image_keys = sorted(self.subdict.keys())
        image_names = [self.subdict[k]["filename"] for k in image_keys]

        select_model = {
            "blip2": MultimodalSearch.load_feature_extractor_model_blip2,
            "blip": MultimodalSearch.load_feature_extractor_model_blip,
            "albef": MultimodalSearch.load_feature_extractor_model_albef,
            "clip_base": MultimodalSearch.load_feature_extractor_model_clip_base,
            "clip_vitl14": MultimodalSearch.load_feature_extractor_model_clip_vitl14,
            "clip_vitl14_336": MultimodalSearch.load_feature_extractor_model_clip_vitl14_336,
        }

        select_extract_image_features = {
            "blip2": MultimodalSearch.extract_image_features_blip2,
            "blip": MultimodalSearch.extract_image_features_basic,
            "albef": MultimodalSearch.extract_image_features_basic,
            "clip_base": MultimodalSearch.extract_image_features_clip,
            "clip_vitl14": MultimodalSearch.extract_image_features_clip,
            "clip_vitl14_336": MultimodalSearch.extract_image_features_clip,
        }

        if model_type in select_model.keys():
            (
                model,
                vis_processors,
                txt_processors,
            ) = select_model[
                model_type
            ](self, MultimodalSearch.multimodal_device)
        else:
            raise SyntaxError(
                "Please, use one of the following models: blip2, blip, albef, clip_base, clip_vitl14, clip_vitl14_336"
            )

        _, images_tensors = MultimodalSearch.read_and_process_images(
            self, image_names, vis_processors
        )
        if path_to_load_tensors is None:
            with torch.no_grad():
                features_image_stacked = select_extract_image_features[model_type](
                    self, model, images_tensors
                )
            MultimodalSearch.save_tensors(
                self, model_type, features_image_stacked, path=path_to_save_tensors
            )
        else:
            features_image_stacked = MultimodalSearch.load_tensors(
                self, str(path_to_load_tensors)
            )

        return (
            model,
            vis_processors,
            txt_processors,
            image_keys,
            image_names,
            features_image_stacked,
        )

    def querys_processing(
        self, search_query: list, model, txt_processors, vis_processors, model_type: str
    ) -> torch.Tensor:
        """
        Process querys.

        Args:
            search_query (list): list of querys.
            model (torch.nn.Module): model.
            txt_processors (dict): preprocessors for text inputs.
            vis_processors (dict): preprocessors for visual inputs.
            model_type (str): type of the model.

        Returns:
            multi_features_stacked (torch.Tensor): tensors of querys features.
        """
        select_extract_image_features = {
            "blip2": MultimodalSearch.extract_image_features_blip2,
            "blip": MultimodalSearch.extract_image_features_basic,
            "albef": MultimodalSearch.extract_image_features_basic,
            "clip_base": MultimodalSearch.extract_image_features_clip,
            "clip_vitl14": MultimodalSearch.extract_image_features_clip,
            "clip_vitl14_336": MultimodalSearch.extract_image_features_clip,
        }

        for query in search_query:
            if len(query) != 1 and (query in ("image", "text_input")):
                raise SyntaxError(
                    'Each query must contain either an "image" or a "text_input"'
                )
        multi_sample = []
        for query in search_query:
            if "text_input" in query.keys():
                text_processing = txt_processors["eval"](query["text_input"])
                images_tensors = ""
            elif "image" in query.keys():
                _, images_tensors = MultimodalSearch.read_and_process_images(
                    self, [query["image"]], vis_processors
                )
                text_processing = ""
            multi_sample.append(
                {"image": images_tensors, "text_input": text_processing}
            )

        multi_features_query = []
        for query in multi_sample:
            if query["image"] == "":
                if model_type in ("clip_base", "clip_vitl14_336", "clip_vitl14"):
                    features = model.extract_features(
                        {"text_input": query["text_input"]}
                    )

                    features_squeeze = features.squeeze(0).to(
                        MultimodalSearch.multimodal_device
                    )
                    multi_features_query.append(
                        Func.normalize(features_squeeze, dim=-1)
                    )
                else:
                    features = model.extract_features(query, mode="text")
                    features_squeeze = (
                        features.text_embeds_proj[:, 0, :]
                        .squeeze(0)
                        .to(MultimodalSearch.multimodal_device)
                    )
                    multi_features_query.append(features_squeeze)
            if query["text_input"] == "":
                multi_features_query.append(
                    select_extract_image_features[model_type](
                        self, model, query["image"]
                    )
                )

        multi_features_stacked = torch.stack(
            [query.squeeze(0) for query in multi_features_query]
        ).to(MultimodalSearch.multimodal_device)

        return multi_features_stacked

    def multimodal_search(
        self,
        model,
        vis_processors,
        txt_processors,
        model_type: str,
        image_keys: list,
        features_image_stacked: torch.Tensor,
        search_query: list,
        filter_number_of_images: str = None,
        filter_val_limit: str = None,
        filter_rel_error: str = None,
    ) -> tuple:
        """
        Search for images with given querys.

        Args:
            model (torch.nn.Module): model.
            vis_processors (dict): preprocessors for visual inputs.
            txt_processors (dict): preprocessors for text inputs.
            model_type (str): type of the model.
            image_keys (list): sorted list of image keys.
            features_image_stacked (torch.Tensor): tensors of images features stacked in device.
            search_query (list): list of querys.
            filter_number_of_images (str): number of images to show. Default: None.
            filter_val_limit (str): limit of similarity value. Default: None.
            filter_rel_error (str): limit of relative error. Default: None.

        Returns:
            similarity (torch.Tensor): similarity between images and querys.
            sorted_lists (list): sorted list of similarity.
        """
        if filter_number_of_images is None:
            filter_number_of_images = len(self.subdict)
        if filter_val_limit is None:
            filter_val_limit = 0
        if filter_rel_error is None:
            filter_rel_error = 1e10

        features_image_stacked.to(MultimodalSearch.multimodal_device)

        with torch.no_grad():
            multi_features_stacked = MultimodalSearch.querys_processing(
                self, search_query, model, txt_processors, vis_processors, model_type
            )

        similarity = features_image_stacked @ multi_features_stacked.t()
        sorted_lists = torch.argsort(similarity, dim=0, descending=True).T.tolist()
        places = [[item.index(i) for i in range(len(item))] for item in sorted_lists]

        for q in range(len(search_query)):
            max_val = similarity[sorted_lists[q][0]][q].item()
            for i, key in zip(range(len(image_keys)), sorted_lists[q]):
                if (
                    i < filter_number_of_images
                    and similarity[key][q].item() > filter_val_limit
                    and 100 * abs(max_val - similarity[key][q].item()) / max_val
                    < filter_rel_error
                ):
                    self.subdict[image_keys[key]][
                        "rank " + list(search_query[q].values())[0]
                    ] = places[q][key]
                    self.subdict[image_keys[key]][
                        list(search_query[q].values())[0]
                    ] = similarity[key][q].item()
                else:
                    self.subdict[image_keys[key]][
                        "rank " + list(search_query[q].values())[0]
                    ] = None
                    self.subdict[image_keys[key]][list(search_query[q].values())[0]] = 0
        return similarity, sorted_lists

    def itm_text_precessing(self, search_query: list[dict[str, str]]) -> list:
        """
        Process text querys for itm model.

        Args:
            search_query (list): list of querys.

        Returns:
            text_query_index (list): list of indexes of text querys.
        """
        for query in search_query:
            if (len(query) != 1) and (query in ("image", "text_input")):
                raise SyntaxError(
                    'Each querry must contain either an "image" or a "text_input"'
                )
        text_query_index = []
        for i, query in zip(range(len(search_query)), search_query):
            if "text_input" in query.keys():
                text_query_index.append(i)

        return text_query_index

    def get_pathes_from_query(self, query: dict[str, str]) -> tuple:
        """
        Get pathes and image names from query.

        Args:
            query (dict): query.

        Returns:
            paths (list): list of pathes.
            image_names (list): list of image names.
        """
        paths = []
        image_names = []
        for s in sorted(
            self.subdict.items(),
            key=lambda t: t[1][list(query.values())[0]],
            reverse=True,
        ):
            if s[1]["rank " + list(query.values())[0]] is None:
                break
            paths.append(s[1]["filename"])
            image_names.append(s[0])
        return paths, image_names

    def read_and_process_images_itm(self, image_paths: list, vis_processor) -> tuple:
        """
        Read and process images with vis_processor for itm model.

        Args:
            image_paths (list): paths to images.
            vis_processor (dict): preprocessors for visual inputs.

        Returns:
            raw_images (list): list of images.
            images_tensors (torch.Tensor): tensors of images stacked in device.
        """
        raw_images = [MultimodalSearch.read_img(self, path) for path in image_paths]
        images = [vis_processor(r_img) for r_img in raw_images]
        images_tensors = torch.stack(images).to(MultimodalSearch.multimodal_device)

        return raw_images, images_tensors

    def compute_gradcam_batch(
        self,
        model: torch.nn.Module,
        visual_input: torch.Tensor,
        text_input: str,
        tokenized_text: torch.Tensor,
        block_num: str = 6,
    ) -> tuple:
        """
        Compute gradcam for itm model.

        Args:
            model (torch.nn.Module): model.
            visual_input (torch.Tensor): tensors of images features stacked in device.
            text_input (str): text.
            tokenized_text (torch.Tensor): tokenized text.
            block_num (int): number of block. Default: 6.

        Returns:
            gradcam (torch.Tensor): gradcam.
            output (torch.Tensor): output of model.
        """
        model.text_encoder.base_model.base_model.encoder.layer[
            block_num
        ].crossattention.self.save_attention = True

        output = model(
            {"image": visual_input, "text_input": text_input}, match_head="itm"
        )
        loss = output[:, 1].sum()

        model.zero_grad()
        loss.backward()
        with torch.no_grad():
            mask = tokenized_text.attention_mask.view(
                tokenized_text.attention_mask.size(0), 1, -1, 1, 1
            )  # (bsz,1,token_len, 1,1)
            token_length = mask.sum() - 2
            token_length = token_length.cpu()
            # grads and cams [bsz, num_head, seq_len, image_patch]
            grads = model.text_encoder.base_model.base_model.encoder.layer[
                block_num
            ].crossattention.self.get_attn_gradients()
            cams = model.text_encoder.base_model.base_model.encoder.layer[
                block_num
            ].crossattention.self.get_attention_map()

            # assume using vit large with 576 num image patch
            cams = (
                cams[:, :, :, 1:].reshape(visual_input.size(0), 12, -1, 24, 24) * mask
            )
            grads = (
                grads[:, :, :, 1:]
                .clamp(0)
                .reshape(visual_input.size(0), 12, -1, 24, 24)
                * mask
            )

            gradcam = cams * grads
            # [enc token gradcam, average gradcam across token, gradcam for individual token]
            # gradcam = torch.cat((gradcam[0:1,:], gradcam[1:token_length+1, :].sum(dim=0, keepdim=True)/token_length, gradcam[1:, :]))
            gradcam = gradcam.mean(1).cpu().detach()
            gradcam = (
                gradcam[:, 1 : token_length + 1, :].sum(dim=1, keepdim=True)
                / token_length
            )

        return gradcam, output

    def resize_img(self, raw_img: Image):
        """
        Proportional resize image to 240 p width.

        Args:
            raw_img (PIL.Image): image.

        Returns:
            resized_image (PIL.Image): proportional resized image to 240p.
        """
        w, h = raw_img.size
        scaling_factor = 240 / w
        resized_image = raw_img.resize(
            (int(w * scaling_factor), int(h * scaling_factor))
        )
        return resized_image

    def get_att_map(
        self,
        img: np.ndarray,
        att_map: np.ndarray,
        blur: bool = True,
        overlap: bool = True,
    ) -> np.ndarray:
        """
        Get attention map.

        Args:
            img (np.ndarray): image.
            att_map (np.ndarray): attention map.
            blur (bool): blur attention map. Default: True.
            overlap (bool): overlap attention map with image. Default: True.

        Returns:
            att_map (np.ndarray): attention map.
        """
        att_map -= att_map.min()
        if att_map.max() > 0:
            att_map /= att_map.max()
        att_map = skimage_transform.resize(
            att_map, (img.shape[:2]), order=3, mode="constant"
        )
        if blur:
            att_map = filters.gaussian_filter(att_map, 0.02 * max(img.shape[:2]))
            att_map -= att_map.min()
            att_map /= att_map.max()
        cmap = plt.get_cmap("jet")
        att_mapv = cmap(att_map)
        att_mapv = np.delete(att_mapv, 3, 2)
        if overlap:
            att_map = (
                1 * (1 - att_map**0.7).reshape(att_map.shape + (1,)) * img
                + (att_map**0.7).reshape(att_map.shape + (1,)) * att_mapv
            )
        return att_map

    def upload_model_blip2_coco(self) -> tuple:
        """
        Load coco blip2_image_text_matching model and preprocessors for visual inputs from lavis.models.

        Returns:
            itm_model (torch.nn.Module): model.
            vis_processor (dict): preprocessors for visual inputs.
        """
        itm_model = load_model(
            "blip2_image_text_matching",
            "coco",
            is_eval=True,
            device=MultimodalSearch.multimodal_device,
        )
        vis_processor = load_processor("blip_image_eval").build(image_size=364)
        return itm_model, vis_processor

    def upload_model_blip_base(self) -> tuple:
        """
        Load base blip_image_text_matching model and preprocessors for visual input from lavis.models.

        Returns:
            itm_model (torch.nn.Module): model.
            vis_processor (dict): preprocessors for visual inputs.
        """
        itm_model = load_model(
            "blip_image_text_matching",
            "base",
            is_eval=True,
            device=MultimodalSearch.multimodal_device,
        )
        vis_processor = load_processor("blip_image_eval").build(image_size=384)
        return itm_model, vis_processor

    def upload_model_blip_large(self) -> tuple:
        """
        Load large blip_image_text_matching model and preprocessors for visual input from lavis.models.

        Returns:
            itm_model (torch.nn.Module): model.
            vis_processor (dict): preprocessors for visual inputs.
        """
        itm_model = load_model(
            "blip_image_text_matching",
            "large",
            is_eval=True,
            device=MultimodalSearch.multimodal_device,
        )
        vis_processor = load_processor("blip_image_eval").build(image_size=384)
        return itm_model, vis_processor

    def image_text_match_reordering(
        self,
        search_query: list[dict[str, str]],
        itm_model_type: str,
        image_keys: list,
        sorted_lists: list[list],
        batch_size: int = 1,
        need_grad_cam: bool = False,
    ) -> tuple:
        """
        Reorder images with itm model.

        Args:
            search_query (list): list of querys.
            itm_model_type (str): type of the model.
            image_keys (list): sorted list of image keys.
            sorted_lists (list): sorted list of similarity.
            batch_size (int): batch size. Default: 1.
            need_grad_cam (bool): need gradcam. Default: False. blip2_coco model does not yet work with gradcam.

        Returns:
            itm_scores2: list of itm scores.
            image_gradcam_with_itm: dict of image names and gradcam.
        """
        if itm_model_type == "blip2_coco" and need_grad_cam is True:
            raise SyntaxError(
                "The blip2_coco model does not yet work with gradcam. Please set need_grad_cam to False"
            )

        choose_model = {
            "blip_base": MultimodalSearch.upload_model_blip_base,
            "blip_large": MultimodalSearch.upload_model_blip_large,
            "blip2_coco": MultimodalSearch.upload_model_blip2_coco,
        }

        itm_model, vis_processor_itm = choose_model[itm_model_type](self)
        text_processor = load_processor("blip_caption")
        tokenizer = BlipBase.init_tokenizer()

        if itm_model_type == "blip2_coco":
            need_grad_cam = False

        text_query_index = MultimodalSearch.itm_text_precessing(self, search_query)

        avg_gradcams = []
        itm_scores = []
        itm_scores2 = []
        image_gradcam_with_itm = {}

        for index_text_query in text_query_index:
            query = search_query[index_text_query]
            pathes, image_names = MultimodalSearch.get_pathes_from_query(self, query)
            num_batches = int(len(pathes) / batch_size)
            num_batches_residue = len(pathes) % batch_size

            local_itm_scores = []
            local_avg_gradcams = []

            if num_batches_residue != 0:
                num_batches = num_batches + 1
            for i in range(num_batches):
                filenames_in_batch = pathes[i * batch_size : (i + 1) * batch_size]
                current_len = len(filenames_in_batch)
                raw_images, images = MultimodalSearch.read_and_process_images_itm(
                    self, filenames_in_batch, vis_processor_itm
                )
                queries_batch = [text_processor(query["text_input"])] * current_len
                queries_tok_batch = tokenizer(queries_batch, return_tensors="pt").to(
                    MultimodalSearch.multimodal_device
                )

                if need_grad_cam:
                    gradcam, itm_output = MultimodalSearch.compute_gradcam_batch(
                        self,
                        itm_model,
                        images,
                        queries_batch,
                        queries_tok_batch,
                    )
                    norm_imgs = [np.float32(r_img) / 255 for r_img in raw_images]

                    for norm_img, grad_cam in zip(norm_imgs, gradcam):
                        avg_gradcam = MultimodalSearch.get_att_map(
                            self, norm_img, np.float32(grad_cam[0]), blur=True
                        )
                        local_avg_gradcams.append(avg_gradcam)

                else:
                    itm_output = itm_model(
                        {"image": images, "text_input": queries_batch}, match_head="itm"
                    )

                with torch.no_grad():
                    itm_score = torch.nn.functional.softmax(itm_output, dim=1)

                local_itm_scores.append(itm_score)

            local_itm_scores2 = torch.cat(local_itm_scores)[:, 1]
            if need_grad_cam:
                localimage_gradcam_with_itm = {
                    n: i * 255 for n, i in zip(image_names, local_avg_gradcams)
                }
            else:
                localimage_gradcam_with_itm = ""
            image_names_with_itm = {
                n: i.item() for n, i in zip(image_names, local_itm_scores2)
            }
            itm_rank = torch.argsort(local_itm_scores2, descending=True)
            image_names_with_new_rank = {
                image_names[i.item()]: rank
                for i, rank in zip(itm_rank, range(len(itm_rank)))
            }
            for i, key in zip(range(len(image_keys)), sorted_lists[index_text_query]):
                if image_keys[key] in image_names:
                    self.subdict[image_keys[key]][
                        "itm " + list(search_query[index_text_query].values())[0]
                    ] = image_names_with_itm[image_keys[key]]
                    self.subdict[image_keys[key]][
                        "itm_rank " + list(search_query[index_text_query].values())[0]
                    ] = image_names_with_new_rank[image_keys[key]]
                else:
                    self.subdict[image_keys[key]][
                        "itm " + list(search_query[index_text_query].values())[0]
                    ] = 0
                    self.subdict[image_keys[key]][
                        "itm_rank " + list(search_query[index_text_query].values())[0]
                    ] = None

            avg_gradcams.append(local_avg_gradcams)
            itm_scores.append(local_itm_scores)
            itm_scores2.append(local_itm_scores2)
            image_gradcam_with_itm[
                list(search_query[index_text_query].values())[0]
            ] = localimage_gradcam_with_itm
        del (
            itm_model,
            vis_processor_itm,
            text_processor,
            raw_images,
            images,
            tokenizer,
            queries_batch,
            queries_tok_batch,
            itm_score,
        )
        if need_grad_cam:
            del itm_output, gradcam, norm_img, grad_cam, avg_gradcam
        torch.cuda.empty_cache()
        return itm_scores2, image_gradcam_with_itm

    def show_results(
        self, query: dict, itm: bool = False, image_gradcam_with_itm: dict = {}
    ) -> None:
        """
        Show results of search.

        Args:
            query (dict): query.
            itm (bool): use itm model. Default: False.
            image_gradcam_with_itm (dict): use gradcam. Default: empty.
        """
        if "image" in query.keys():
            pic = Image.open(query["image"]).convert("RGB")
            pic.thumbnail((400, 400))
            display(
                "Your search query: ",
                pic,
                "--------------------------------------------------",
                "Results:",
            )
        elif "text_input" in query.keys():
            display(
                "Your search query: " + query["text_input"],
                "--------------------------------------------------",
                "Results:",
            )
        if itm:
            current_querry_val = "itm " + list(query.values())[0]
            current_querry_rank = "itm_rank " + list(query.values())[0]
        else:
            current_querry_val = list(query.values())[0]
            current_querry_rank = "rank " + list(query.values())[0]

        for s in sorted(
            self.subdict.items(), key=lambda t: t[1][current_querry_val], reverse=True
        ):
            if s[1][current_querry_rank] is None:
                break
            if bool(image_gradcam_with_itm) is True and itm is True:
                image = image_gradcam_with_itm[list(query.values())[0]][s[0]]
                p1 = Image.fromarray(image.astype("uint8"), "RGB")
            else:
                p1 = Image.open(s[1]["filename"]).convert("RGB")
            p1.thumbnail((400, 400))
            display(
                "Rank: "
                + str(s[1][current_querry_rank])
                + " Val: "
                + str(s[1][current_querry_val]),
                s[0],
                p1,
            )
            display(
                "--------------------------------------------------",
            )



try:
    from importlib import metadata
except ImportError:
    # Running on pre-3.8 Python; use importlib-metadata package
    import importlib_metadata as metadata  # type: ignore
from ammico.cropposts import crop_media_posts, crop_posts_from_refs
from ammico.display import AnalysisExplorer
from ammico.faces import EmotionDetector
from ammico.multimodal_search import MultimodalSearch
from ammico.objects import ObjectDetector
from ammico.summary import SummaryDetector
from ammico.text import TextDetector, PostprocessText
from ammico.utils import find_files, get_dataframe

# Export the version defined in project metadata
__version__ = metadata.version(__package__)
del metadata

__all__ = [
    "crop_media_posts",
    "crop_posts_from_refs",
    "AnalysisExplorer",
    "EmotionDetector",
    "MultimodalSearch",
    "ObjectDetector",
    "SummaryDetector",
    "TextDetector",
    "PostprocessText",
    "find_files",
    "get_dataframe",
]



import numpy as np
import webcolors
import pandas as pd
from collections import defaultdict
import colorgram
import colour
from ammico.utils import get_color_table, AnalysisMethod

COLOR_SCHEMES = [
    "CIE 1976",
    "CIE 1994",
    "CIE 2000",
    "CMC",
    "ITP",
    "CAM02-LCD",
    "CAM02-SCD",
    "CAM02-UCS",
    "CAM16-LCD",
    "CAM16-SCD",
    "CAM16-UCS",
    "DIN99",
]


class ColorDetector(AnalysisMethod):
    def __init__(
        self,
        subdict: dict,
        delta_e_method: str = "CIE 1976",
    ) -> None:
        """Color Analysis class, analyse hue and identify named colors.

        Args:
            subdict (dict): The dictionary containing the image path.
            delta_e_method (str): The calculation method used for assigning the
                closest color name, defaults to "CIE 1976".
                The available options are: 'CIE 1976', 'CIE 1994', 'CIE 2000',
                'CMC', 'ITP', 'CAM02-LCD', 'CAM02-SCD', 'CAM02-UCS', 'CAM16-LCD',
                'CAM16-SCD', 'CAM16-UCS', 'DIN99'
        """
        super().__init__(subdict)
        self.subdict.update(self.set_keys())
        self.merge_color = True
        self.n_colors = 100
        if delta_e_method not in COLOR_SCHEMES:
            raise ValueError(
                "Invalid selection for assigning the color name. Please select one of {}".format(
                    COLOR_SCHEMES
                )
            )
        self.delta_e_method = delta_e_method

    def set_keys(self) -> dict:
        colors = {
            "red": 0,
            "green": 0,
            "blue": 0,
            "yellow": 0,
            "cyan": 0,
            "orange": 0,
            "purple": 0,
            "pink": 0,
            "brown": 0,
            "grey": 0,
            "white": 0,
            "black": 0,
        }
        return colors

    def analyse_image(self):
        """
        Uses the colorgram library to extract the n most common colors from the images.
        One problem is, that the most common colors are taken before beeing categorized,
        so for small values it might occur that the ten most common colors are shades of grey,
        while other colors are present but will be ignored. Because of this n_colors=100 was chosen as default.

        The colors are then matched to the closest color in the CSS3 color list using the delta-e metric.
        They are then merged into one data frame.
        The colors can be reduced to a smaller list of colors using the get_color_table function.
        These colors are: "red", "green", "blue", "yellow","cyan", "orange", "purple", "pink", "brown", "grey", "white", "black".

        Returns:
            dict: Dictionary with color names as keys and percentage of color in image as values.
        """
        filename = self.subdict["filename"]

        colors = colorgram.extract(filename, self.n_colors)
        for color in colors:
            rgb_name = self.rgb2name(
                color.rgb,
                merge_color=self.merge_color,
                delta_e_method=self.delta_e_method,
            )
            self.subdict[rgb_name] += color.proportion

        # ensure color rounding
        for key in self.set_keys().keys():
            if self.subdict[key]:
                self.subdict[key] = round(self.subdict[key], 2)

        return self.subdict

    def rgb2name(
        self, c, merge_color: bool = True, delta_e_method: str = "CIE 1976"
    ) -> str:
        """Take an rgb color as input and return the closest color name from the CSS3 color list.

        Args:
            c (Union[List,tuple]): RGB value.
            merge_color (bool, Optional): Whether color name should be reduced, defaults to True.
        Returns:
            str: Closest matching color name.
        """
        if len(c) != 3:
            raise ValueError("Input color must be a list or tuple of length 3 (RGB).")

        h_color = "#{:02x}{:02x}{:02x}".format(int(c[0]), int(c[1]), int(c[2]))
        try:
            output_color = webcolors.hex_to_name(h_color, spec="css3")
            output_color = output_color.lower().replace("grey", "gray")
        except ValueError:
            delta_e_lst = []
            filtered_colors = webcolors.CSS3_NAMES_TO_HEX

            for _, img_hex in filtered_colors.items():
                cur_clr = webcolors.hex_to_rgb(img_hex)
                # calculate color Delta-E
                delta_e = colour.delta_E(c, cur_clr, method=delta_e_method)
                delta_e_lst.append(delta_e)
            # find lowest delta-e
            min_diff = np.argsort(delta_e_lst)[0]
            output_color = (
                str(list(filtered_colors.items())[min_diff][0])
                .lower()
                .replace("grey", "gray")
            )

        # match color to reduced list:
        if merge_color:
            for reduced_key, reduced_color_sub_list in get_color_table().items():
                if str(output_color).lower() in [
                    str(color_name).lower()
                    for color_name in reduced_color_sub_list["ColorName"]
                ]:
                    output_color = reduced_key.lower()
                    break
        return output_color



import ammico.faces as faces
import ammico.text as text
import ammico.objects as objects
import ammico.colors as colors
from ammico.utils import is_interactive
import ammico.summary as summary
import dash_renderjson
from dash import html, Input, Output, dcc, State, Dash
from PIL import Image
import dash_bootstrap_components as dbc


COLOR_SCHEMES = [
    "CIE 1976",
    "CIE 1994",
    "CIE 2000",
    "CMC",
    "ITP",
    "CAM02-LCD",
    "CAM02-SCD",
    "CAM02-UCS",
    "CAM16-LCD",
    "CAM16-SCD",
    "CAM16-UCS",
    "DIN99",
]
SUMMARY_ANALYSIS_TYPE = ["summary_and_questions", "summary", "questions"]
SUMMARY_MODEL = ["base", "large"]


class AnalysisExplorer:
    def __init__(self, mydict: dict) -> None:
        """Initialize the AnalysisExplorer class to create an interactive
        visualization of the analysis results.

        Args:
            mydict (dict): A nested dictionary containing image data for all images.

        """
        self.app = Dash(__name__, external_stylesheets=[dbc.themes.BOOTSTRAP])
        self.mydict = mydict
        self.theme = {
            "scheme": "monokai",
            "author": "wimer hazenberg (http://www.monokai.nl)",
            "base00": "#272822",
            "base01": "#383830",
            "base02": "#49483e",
            "base03": "#75715e",
            "base04": "#a59f85",
            "base05": "#f8f8f2",
            "base06": "#f5f4f1",
            "base07": "#f9f8f5",
            "base08": "#f92672",
            "base09": "#fd971f",
            "base0A": "#f4bf75",
            "base0B": "#a6e22e",
            "base0C": "#a1efe4",
            "base0D": "#66d9ef",
            "base0E": "#ae81ff",
            "base0F": "#cc6633",
        }

        # Setup the layout
        app_layout = html.Div(
            [
                # Top row, only file explorer
                dbc.Row(
                    [dbc.Col(self._top_file_explorer(mydict))],
                    id="Div_top",
                    style={
                        "width": "30%",
                    },
                ),
                # second row, middle picture and right output
                dbc.Row(
                    [
                        # first column: picture
                        dbc.Col(self._middle_picture_frame()),
                        dbc.Col(self._right_output_json()),
                    ]
                ),
            ],
            # style={"width": "95%", "display": "inline-block"},
        )
        self.app.layout = app_layout

        # Add callbacks to the app
        self.app.callback(
            Output("img_middle_picture_id", "src"),
            Input("left_select_id", "value"),
            prevent_initial_call=True,
        )(self.update_picture)

        self.app.callback(
            Output("right_json_viewer", "data"),
            Input("button_run", "n_clicks"),
            State("left_select_id", "options"),
            State("left_select_id", "value"),
            State("Dropdown_select_Detector", "value"),
            State("setting_Text_analyse_text", "value"),
            State("setting_Text_model_names", "value"),
            State("setting_Text_revision_numbers", "value"),
            State("setting_Emotion_emotion_threshold", "value"),
            State("setting_Emotion_race_threshold", "value"),
            State("setting_Color_delta_e_method", "value"),
            State("setting_Summary_analysis_type", "value"),
            State("setting_Summary_model", "value"),
            State("setting_Summary_list_of_questions", "value"),
            prevent_initial_call=True,
        )(self._right_output_analysis)

        self.app.callback(
            Output("settings_TextDetector", "style"),
            Output("settings_EmotionDetector", "style"),
            Output("settings_ColorDetector", "style"),
            Output("settings_Summary_Detector", "style"),
            Input("Dropdown_select_Detector", "value"),
        )(self._update_detector_setting)

    # I split the different sections into subfunctions for better clarity
    def _top_file_explorer(self, mydict: dict) -> html.Div:
        """Initialize the file explorer dropdown for selecting the file to be analyzed.

        Args:
            mydict (dict): A dictionary containing image data.

        Returns:
            html.Div: The layout for the file explorer dropdown.
        """
        left_layout = html.Div(
            [
                dcc.Dropdown(
                    options={value["filename"]: key for key, value in mydict.items()},
                    id="left_select_id",
                )
            ]
        )
        return left_layout

    def _middle_picture_frame(self) -> html.Div:
        """Initialize the picture frame to display the image.

        Returns:
            html.Div: The layout for the picture frame.
        """
        middle_layout = html.Div(
            [
                html.Img(
                    id="img_middle_picture_id",
                    style={
                        "width": "80%",
                    },
                )
            ]
        )
        return middle_layout

    def _create_setting_layout(self):
        settings_layout = html.Div(
            [
                # text summary start
                html.Div(
                    id="settings_TextDetector",
                    style={"display": "none"},
                    children=[
                        dbc.Row(
                            dcc.Checklist(
                                ["Analyse text"],
                                ["Analyse text"],
                                id="setting_Text_analyse_text",
                                style={"margin-bottom": "10px"},
                            ),
                        ),  # row 1
                        # text row 2
                        dbc.Row(
                            [
                                dbc.Col(
                                    [
                                        html.P(
                                            "Select models for text_summary, text_sentiment, text_NER or leave blank for default:",
                                            # style={"width": "45%"},
                                        ),
                                    ]
                                ),  #
                                dbc.Col(
                                    [
                                        html.P(
                                            "Select model revision number for text_summary, text_sentiment, text_NER or leave blank for default:"
                                        ),
                                    ]
                                ),
                            ]
                        ),  # row 2
                        #  input row 3
                        dbc.Row(
                            [
                                dbc.Col(
                                    dcc.Input(
                                        type="text",
                                        id="setting_Text_model_names",
                                        style={"width": "100%"},
                                    ),
                                ),
                                dbc.Col(
                                    dcc.Input(
                                        type="text",
                                        id="setting_Text_revision_numbers",
                                        style={"width": "100%"},
                                    ),
                                ),
                            ]
                        ),  # row 3
                    ],
                ),  # text summary end
                # start emotion detector
                html.Div(
                    id="settings_EmotionDetector",
                    style={"display": "none"},
                    children=[
                        dbc.Row(
                            [
                                dbc.Col(
                                    [
                                        html.P("Emotion threshold"),
                                        dcc.Input(
                                            value=50,
                                            type="number",
                                            max=100,
                                            min=0,
                                            id="setting_Emotion_emotion_threshold",
                                            style={"width": "100%"},
                                        ),
                                    ],
                                    align="start",
                                ),
                                dbc.Col(
                                    [
                                        html.P("Race threshold"),
                                        dcc.Input(
                                            type="number",
                                            value=50,
                                            max=100,
                                            min=0,
                                            id="setting_Emotion_race_threshold",
                                            style={"width": "100%"},
                                        ),
                                    ],
                                    align="start",
                                ),
                            ],
                            style={"width": "100%"},
                        ),
                    ],
                ),  # end emotion detector
                html.Div(
                    id="settings_ColorDetector",
                    style={"display": "none"},
                    children=[
                        html.Div(
                            [
                                dcc.Dropdown(
                                    options=COLOR_SCHEMES,
                                    value="CIE 1976",
                                    id="setting_Color_delta_e_method",
                                )
                            ],
                            style={
                                "width": "49%",
                                "display": "inline-block",
                                "margin-top": "10px",
                            },
                        )
                    ],
                ),
                html.Div(
                    id="settings_Summary_Detector",
                    style={"display": "none"},
                    children=[
                        dbc.Col(
                            [
                                dbc.Row([html.P("Analysis type:")]),
                                dbc.Row([html.P("Model type:")]),
                                dbc.Row([html.P("Analysis question:")]),
                            ],
                        ),
                        dbc.Col(
                            [
                                dbc.Row(
                                    dcc.Dropdown(
                                        options=SUMMARY_ANALYSIS_TYPE,
                                        value="summary_and_questions",
                                        id="setting_Summary_analysis_type",
                                    )
                                ),
                                dbc.Row(
                                    dcc.Dropdown(
                                        options=SUMMARY_MODEL,
                                        value="base",
                                        id="setting_Summary_model",
                                    )
                                ),
                                dbc.Row(
                                    dcc.Input(
                                        type="text",
                                        id="setting_Summary_list_of_questions",
                                        style={
                                            "height": "auto",
                                            "margin-left": "11px",
                                        },
                                    ),
                                ),
                            ]
                        ),
                    ],
                ),
            ],
            style={"width": "100%", "display": "inline-block"},
        )
        return settings_layout

    def _right_output_json(self) -> html.Div:
        """Initialize the DetectorDropdown, argument Div and JSON viewer for displaying the analysis output.

        Returns:
            html.Div: The layout for the JSON viewer.
        """
        right_layout = html.Div(
            [
                dbc.Col(
                    [
                        dbc.Row(
                            dcc.Dropdown(
                                options=[
                                    "TextDetector",
                                    "ObjectDetector",
                                    "EmotionDetector",
                                    "SummaryDetector",
                                    "ColorDetector",
                                ],
                                value="TextDetector",
                                id="Dropdown_select_Detector",
                                style={"width": "60%"},
                            ),
                            justify="start",
                        ),
                        dbc.Row(
                            children=[self._create_setting_layout()],
                            id="div_detector_args",
                            justify="start",
                        ),
                        dbc.Row(
                            html.Button(
                                "Run Detector",
                                id="button_run",
                                style={
                                    "margin-top": "15px",
                                    "margin-bottom": "15px",
                                    "margin-left": "11px",
                                    "width": "30%",
                                },
                            ),
                            justify="start",
                        ),
                        dbc.Row(
                            dcc.Loading(
                                id="loading-2",
                                children=[
                                    dash_renderjson.DashRenderjson(
                                        id="right_json_viewer",
                                        data={},
                                        max_depth=-1,
                                        theme=self.theme,
                                        invert_theme=True,
                                    ),
                                ],
                                type="circle",
                            ),
                            justify="start",
                        ),
                    ],
                    align="start",
                )
            ]
        )
        return right_layout

    def run_server(self, port: int = 8050) -> None:
        """Run the Dash server to start the analysis explorer.


        Args:
            port (int, optional): The port number to run the server on (default: 8050).
        """

        self.app.run_server(debug=True, port=port)

    # Dash callbacks
    def update_picture(self, img_path: str):
        """Callback function to update the displayed image.

        Args:
            img_path (str): The path of the selected image.

        Returns:
            Union[PIL.PngImagePlugin, None]: The image object to be displayed
                or None if the image path is

        """
        if img_path is not None:
            image = Image.open(img_path)
            return image
        else:
            return None

    def _update_detector_setting(self, setting_input):
        # return settings_TextDetector -> style, settings_EmotionDetector -> style
        display_none = {"display": "none"}
        display_flex = {
            "display": "flex",
            "flexWrap": "wrap",
            "width": 400,
            "margin-top": "20px",
        }

        if setting_input == "TextDetector":
            return display_flex, display_none, display_none, display_none

        if setting_input == "EmotionDetector":
            return display_none, display_flex, display_none, display_none

        if setting_input == "ColorDetector":
            return display_none, display_none, display_flex, display_none

        if setting_input == "SummaryDetector":
            return display_none, display_none, display_none, display_flex

        else:
            return display_none, display_none, display_none, display_none

    def _right_output_analysis(
        self,
        n_clicks,
        all_img_options: dict,
        current_img_value: str,
        detector_value: str,
        settings_text_analyse_text: list,
        settings_text_model_names: str,
        settings_text_revision_numbers: str,
        setting_emotion_emotion_threshold: int,
        setting_emotion_race_threshold: int,
        setting_color_delta_e_method: str,
        setting_summary_analysis_type: str,
        setting_summary_model: str,
        setting_summary_list_of_questions: str,
    ) -> dict:
        """Callback function to perform analysis on the selected image and return the output.

        Args:
            all_options (dict): The available options in the file explorer dropdown.
            current_value (str): The current selected value in the file explorer dropdown.

        Returns:
            dict: The analysis output for the selected image.
        """
        identify_dict = {
            "EmotionDetector": faces.EmotionDetector,
            "TextDetector": text.TextDetector,
            "ObjectDetector": objects.ObjectDetector,
            "SummaryDetector": summary.SummaryDetector,
            "ColorDetector": colors.ColorDetector,
        }

        # Get image ID from dropdown value, which is the filepath
        if current_img_value is None:
            return {}
        image_id = all_img_options[current_img_value]
        # copy image so prvious runs don't leave their default values in the dict
        image_copy = self.mydict[image_id].copy()

        # detector value is the string name of the chosen detector
        identify_function = identify_dict[detector_value]

        if detector_value == "TextDetector":
            analyse_text = (
                True if settings_text_analyse_text == ["Analyse text"] else False
            )
            detector_class = identify_function(
                image_copy,
                analyse_text=analyse_text,
                model_names=[settings_text_model_names]
                if (settings_text_model_names is not None)
                else None,
                revision_numbers=[settings_text_revision_numbers]
                if (settings_text_revision_numbers is not None)
                else None,
            )
        elif detector_value == "EmotionDetector":
            print("test")
            detector_class = identify_function(
                image_copy,
                race_threshold=setting_emotion_race_threshold,
                emotion_threshold=setting_emotion_emotion_threshold,
            )
        elif detector_value == "ColorDetector":
            detector_class = identify_function(
                image_copy,
                delta_e_method=setting_color_delta_e_method,
            )
        elif detector_value == "SummaryDetector":
            detector_class = identify_function(
                image_copy,
                analysis_type=setting_summary_analysis_type,
                summary_model_type=setting_summary_model,
                list_of_questions=[setting_summary_list_of_questions]
                if (setting_summary_list_of_questions is not None)
                else None,
            )
        else:
            detector_class = identify_function(image_copy)
        return detector_class.analyse_image()
